import asyncio
import contextlib

import pytest
import zmq
from pytest_mock import MockerFixture

from mrok.proxy.events import EventsPublisher
from mrok.proxy.models import (
    DataTransferMetrics,
    Event,
    HTTPHeaders,
    HTTPRequest,
    HTTPResponse,
    Identity,
    ProcessMetrics,
    RequestsMetrics,
    ResponseTimeMetrics,
    Status,
    WorkerMetrics,
)


async def test_publish_metrics_event(
    mocker: MockerFixture,
    ziti_identity_file: str,
):
    identity = Identity.load_from_file(ziti_identity_file)
    event_publisher = EventsPublisher(
        worker_id="my-worker-id",
        meta=identity.mrok,
    )

    metrics_snapshot = WorkerMetrics(
        worker_id="my-worker-id",
        data_transfer=DataTransferMetrics(
            bytes_in=1000,
            bytes_out=2000,
        ),
        requests=RequestsMetrics(rps=123, total=1000, successful=10, failed=30),
        response_time=ResponseTimeMetrics(
            avg=10,
            min=1,
            max=30,
            p50=11,
            p90=22,
            p99=11,
        ),
        process=ProcessMetrics(cpu=12, mem=22),
    )

    event_publisher._publisher = mocker.AsyncMock()
    event_publisher._metrics_collector = mocker.AsyncMock()
    event_publisher._metrics_collector.snapshot.return_value = metrics_snapshot  # type: ignore

    task = asyncio.create_task(event_publisher.publish_metrics_event())
    await asyncio.sleep(0.1)
    task.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await task

    metrics_events = Event(
        type="status",
        data=Status(
            meta=identity.mrok,
            metrics=metrics_snapshot,
        ),
    )
    event_publisher._publisher.send_string.assert_called_once_with(  # type: ignore
        metrics_events.model_dump_json()
    )


@pytest.mark.asyncio
async def test_publish_response_event(
    mocker: MockerFixture,
    ziti_identity_file: str,
):
    identity = Identity.load_from_file(ziti_identity_file)
    event_publisher = EventsPublisher(
        worker_id="my-worker-id",
        meta=identity.mrok,
    )

    resp = HTTPResponse(
        type="response",
        headers=HTTPHeaders.from_asgi([(b"content-type", b"text/plain")]),
        request=HTTPRequest(
            method="GET",
            url="url",
            query_string=b"",
            headers=HTTPHeaders.from_asgi([(b"content-type", b"application/json")]),
            start_time=0,
        ),
        status=200,
        duration=20.5,
    )

    event_publisher._publisher = mocker.AsyncMock()

    await event_publisher.publish_response_event(resp)

    resp_event = Event(type="response", data=resp)
    event_publisher._publisher.send_string.assert_awaited_once_with(  # type: ignore
        resp_event.model_dump_json()
    )


@pytest.mark.asyncio
async def test_lifespan(
    mocker: MockerFixture,
    ziti_identity_file: str,
):
    m_on_startup = mocker.patch.object(EventsPublisher, "on_startup")
    m_on_shutdown = mocker.patch.object(EventsPublisher, "on_shutdown")

    identity = Identity.load_from_file(ziti_identity_file)
    event_publisher = EventsPublisher(
        worker_id="my-worker-id",
        meta=identity.mrok,
    )

    m_app = mocker.AsyncMock()
    async with event_publisher.lifespan(m_app):
        pass

    m_on_startup.assert_awaited_once()
    m_on_shutdown.assert_awaited_once()


@pytest.mark.asyncio
async def test_on_startup(
    mocker: MockerFixture,
    ziti_identity_file: str,
):
    m_publisher = mocker.MagicMock()
    m_zmq_ctx = mocker.MagicMock()
    m_zmq_ctx.socket.return_value = m_publisher
    m_zmq_ctx_ctor = mocker.MagicMock()
    m_zmq_ctx_ctor.return_value = m_zmq_ctx
    mocker.patch("mrok.proxy.events.zmq.asyncio.Context", m_zmq_ctx_ctor)

    m_metrics = mocker.MagicMock()
    m_metricscollector_ctor = mocker.patch(
        "mrok.proxy.events.MetricsCollector", return_value=m_metrics
    )
    m_publish_metrics_event = mocker.patch.object(EventsPublisher, "publish_metrics_event")

    identity = Identity.load_from_file(ziti_identity_file)
    event_publisher = EventsPublisher(
        worker_id="my-worker-id",
        meta=identity.mrok,
        events_publisher_port=8282,
    )

    await event_publisher.on_startup()
    m_metricscollector_ctor.assert_called_once_with("my-worker-id")
    assert event_publisher._metrics_collector == m_metrics
    assert event_publisher._zmq_ctx == m_zmq_ctx
    m_zmq_ctx.socket.assert_called_once_with(zmq.PUB)
    m_publisher.connect.assert_called_once_with("tcp://localhost:8282")
    await asyncio.sleep(0.001)
    m_publish_metrics_event.assert_awaited_once()


@pytest.mark.asyncio
async def test_on_shutdown(
    mocker: MockerFixture,
    ziti_identity_file: str,
):
    identity = Identity.load_from_file(ziti_identity_file)
    event_publisher = EventsPublisher(
        worker_id="my-worker-id",
        meta=identity.mrok,
    )

    async def my_coro():
        while True:
            await asyncio.sleep(5)

    task = asyncio.create_task(my_coro())

    event_publisher._publish_task = task  # type: ignore
    event_publisher._publisher = mocker.MagicMock()
    event_publisher._zmq_ctx = mocker.MagicMock()

    await event_publisher.on_shutdown()
    assert task.cancelled()
    event_publisher._publisher.close.assert_called_once()  # type: ignore
    event_publisher._zmq_ctx.term.assert_called_once()  # type: ignore



from http import HTTPStatus

from httpcore import ConnectError


class ProxyError(Exception):
    def __init__(self, http_status: HTTPStatus, message: str) -> None:
        self.http_status: HTTPStatus = http_status
        self.message: str = message


class InvalidTargetError(ProxyError):
    def __init__(self):
        super().__init__(HTTPStatus.BAD_GATEWAY, "Bad Gateway: invalid target extension.")


class TargetUnavailableError(ProxyError, ConnectError):
    def __init__(self):
        super().__init__(
            HTTPStatus.SERVICE_UNAVAILABLE,
            "Service Unavailable: the target extension is unavailable.",
        )



import re

from mrok.conf import get_settings


def parse_accept_header(accept: str | None) -> list[str]:
    if not accept:
        return ["*/*"]

    result: list[tuple[str, float, int]] = []

    for index, item in enumerate(accept.split(",")):
        item = item.strip()
        if not item:
            continue

        parts = [p.strip() for p in item.split(";")]
        media_type = parts[0].lower()

        q = 1.0
        for param in parts[1:]:
            if param.startswith("q="):
                try:
                    q = float(param[2:])
                except ValueError:
                    q = 0.0

        result.append((media_type, q, index))

    # Sort by:
    # 1) q value (desc)
    # 2) specificity (more specific first)
    # 3) original order (stable)
    result.sort(
        key=lambda x: (
            -x[1],
            -_media_type_specificity(x[0]),
            x[2],
        )
    )

    return [media_type for media_type, _, _ in result]


def _media_type_specificity(media_type: str) -> int:
    if media_type == "*/*":
        return 0
    if media_type.endswith("/*"):
        return 1
    return 2


def get_frontend_domain():
    settings = get_settings()
    return (
        settings.frontend.domain
        if settings.frontend.domain[0] == "."
        else f".{settings.frontend.domain}"
    )


def _get_target_from_header(headers: dict[str, str], name: str) -> str | None:
    domain_name = get_frontend_domain()
    header_value = headers.get(name, "")
    if domain_name in header_value:
        if ":" in header_value:
            header_value, _ = header_value.split(":", 1)
        return header_value[: -len(domain_name)]


def get_target_name(headers: dict[str, str]) -> str | None:
    settings = get_settings()

    target = _get_target_from_header(headers, "x-forwarded-host")
    if not target:
        target = _get_target_from_header(headers, "host")

    if target and (
        re.fullmatch(settings.identifiers.extension.regex, target)
        or re.fullmatch(settings.identifiers.instance.regex, target)
    ):
        return target
    return None



import signal
import threading
import time
from collections.abc import Generator
from pathlib import Path

import zmq
from pytest_mock import MockerFixture
from watchfiles import Change

from mrok.proxy.master import (
    MONITOR_THREAD_JOIN_TIMEOUT,
    MasterBase,
    start_events_router,
    start_uvicorn_worker,
)
from tests.conftest import SettingsFactory


def test_start_events_router_hook(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocker.patch("mrok.proxy.master.get_settings", return_value=settings)
    m_setup_logging = mocker.patch("mrok.proxy.master.setup_logging")
    m_frontend = mocker.MagicMock()
    m_backend = mocker.MagicMock()
    m_zmq_ctx = mocker.MagicMock()
    m_zmq_ctx.socket.side_effect = [m_frontend, m_backend]
    m_zmq_ctx_ctor = mocker.MagicMock()
    m_zmq_ctx_ctor.return_value = m_zmq_ctx
    mocker.patch("mrok.proxy.master.zmq.Context", m_zmq_ctx_ctor)
    m_proxy = mocker.patch("mrok.proxy.master.zmq.proxy")

    start_events_router(5000, 5101)

    m_setup_logging.assert_called_once_with(settings)
    assert m_zmq_ctx.socket.mock_calls[0].args[0] == zmq.XSUB
    assert m_zmq_ctx.socket.mock_calls[1].args[0] == zmq.XPUB
    m_frontend.bind.assert_called_once_with("tcp://localhost:5000")
    m_backend.bind.assert_called_once_with("tcp://localhost:5101")
    m_proxy.assert_called_once_with(m_frontend, m_backend)
    m_frontend.close.assert_called_once()
    m_backend.close.assert_called_once()
    m_zmq_ctx.term.assert_called_once()


def test_start_uvicorn_worker_hook(
    mocker: MockerFixture,
):
    m_worker = mocker.MagicMock()
    m_worker_ctor = mocker.patch(
        "mrok.proxy.master.Worker",
        return_value=m_worker,
    )
    m_app = mocker.MagicMock()
    start_uvicorn_worker(
        "my-wk-id",
        m_app,
        "my-id-file.json",
        events_enabled=False,
        events_pub_port=2233,
        events_metrics_collect_interval=24.0,
    )
    m_worker_ctor.assert_called_once_with(
        "my-wk-id",
        m_app,
        "my-id-file.json",
        events_enabled=False,
        events_publisher_port=2233,
        events_metrics_collect_interval=24.0,
        ziti_load_timeout_ms=5000,
        server_backlog=2048,
        server_timeout_keep_alive=5,
        server_limit_concurrency=None,
        server_limit_max_requests=None,
    )
    m_worker.run.assert_called_once()


def test_init(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    mocked_setup_signals = mocker.patch.object(Master, "setup_signals_handler")

    master = Master(
        "my-identity.json",
        server_workers=3,
        server_reload=True,
        events_enabled=False,
        events_pub_port=50000,
        events_sub_port=51000,
        events_metrics_collect_interval=7.0,
    )
    assert master.identity_file == "my-identity.json"
    assert master.workers == 3
    assert master.server_reload is True
    assert master.events_enabled is False
    assert master.events_pub_port == 50000
    assert master.events_sub_port == 51000
    assert master.events_metrics_collect_interval == 7.0
    mocked_setup_signals.assert_called_once()


def test_setup_signals_handler(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    mocked_signal = mocker.patch("mrok.proxy.master.signal.signal")
    master = Master("my-identity.json")
    assert mocked_signal.call_count == 2
    assert mocked_signal.mock_calls[0].args == (signal.SIGINT, master.handle_signal)
    assert mocked_signal.mock_calls[1].args == (signal.SIGTERM, master.handle_signal)


def test_handle_signal(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    master = Master("my-identity.json")
    assert master.stop_event.is_set() is False
    master.handle_signal()
    assert master.stop_event.is_set() is True


def test_start_worker(mocker: MockerFixture):
    m_asgi_app = mocker.AsyncMock()

    class Master(MasterBase):
        def get_asgi_app(self):
            return m_asgi_app

    m_proc = mocker.MagicMock()
    mocked_start_process = mocker.patch("mrok.proxy.master.start_process", return_value=m_proc)

    master = Master(
        "my-identity.json",
        server_workers=3,
        server_reload=False,
        events_enabled=True,
        events_pub_port=50000,
        events_sub_port=51000,
        events_metrics_collect_interval=10,
    )
    assert master.start_worker("my-worker-id") == m_proc
    mocked_start_process.assert_called_once_with(
        start_uvicorn_worker,
        "function",
        (
            "my-worker-id",
            m_asgi_app,
            "my-identity.json",
        ),
        {
            "ziti_load_timeout_ms": 5000,
            "server_backlog": 2048,
            "server_timeout_keep_alive": 5,
            "server_limit_concurrency": None,
            "server_limit_max_requests": None,
            "events_metrics_collect_interval": 10,
            "events_enabled": True,
            "events_pub_port": 50000,
        },
    )


def test_start_events_router(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    m_proc = mocker.MagicMock()
    mocked_start_process = mocker.patch("mrok.proxy.master.start_process", return_value=m_proc)

    master = Master(
        "my-identity.json",
        events_pub_port=50000,
        events_sub_port=51000,
    )
    master.start_events_router()
    assert master.zmq_pubsub_router_process == m_proc
    mocked_start_process.assert_called_once_with(
        start_events_router,
        "function",
        (50000, 51000),
        None,
    )


def test_start_workers(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    w0 = mocker.MagicMock()
    w1 = mocker.MagicMock()
    w2 = mocker.MagicMock()

    mocked_start_worker = mocker.patch.object(Master, "start_worker", side_effect=[w0, w1, w2])

    master = Master("my-identity.json", server_workers=3)
    master.start_workers()
    assert master.worker_processes[master.worker_identifiers[0]] == w0
    assert master.worker_processes[master.worker_identifiers[1]] == w1
    assert master.worker_processes[master.worker_identifiers[2]] == w2
    for i in range(3):
        assert mocked_start_worker.mock_calls[i].args[0] == master.worker_identifiers[i]


def test_start(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    mocked_start_events_router = mocker.patch.object(Master, "start_events_router")
    mocked_start_workers = mocker.patch.object(Master, "start_workers")
    mocked_monitor_thread = mocker.MagicMock()

    master = Master(
        "my-identity.json",
    )
    master.monitor_thread = mocked_monitor_thread
    master.start()
    mocked_start_events_router.assert_called_once()
    mocked_start_workers.assert_called_once()
    mocked_monitor_thread.start.assert_called_once()


def test_stop_workers(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    w0 = mocker.MagicMock()
    w1 = mocker.MagicMock()
    master = Master("my-identity.json")

    master.worker_processes = {"id1": w0, "id2": w1}
    master.stop_workers()
    w0.stop.assert_called_once_with(sigint_timeout=5, sigkill_timeout=1)
    w1.stop.assert_called_once_with(sigint_timeout=5, sigkill_timeout=1)
    assert master.worker_processes == {}


def test_stop_event_router(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    master = Master("my-identity.json")
    mocked_zmq_pubsub_router_process = mocker.MagicMock()
    master.zmq_pubsub_router_process = mocked_zmq_pubsub_router_process
    master.stop_events_router()
    mocked_zmq_pubsub_router_process.stop.assert_called_once_with(
        sigint_timeout=5, sigkill_timeout=1
    )


def test_stop(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    mocked_stop_events_router = mocker.patch.object(Master, "stop_events_router")
    mocked_stop_workers = mocker.patch.object(Master, "stop_workers")
    mocked_monitor_thread = mocker.MagicMock()
    mocked_monitor_thread.is_alive.return_value = True

    master = Master("my-identity.json")
    master.monitor_thread = mocked_monitor_thread
    master.stop()

    mocked_stop_events_router.assert_called_once()
    mocked_stop_workers.assert_called_once()
    mocked_monitor_thread.join.assert_called_once_with(timeout=MONITOR_THREAD_JOIN_TIMEOUT)


def test_restart(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    mocked_stop = mocker.patch.object(Master, "stop_workers")
    mocked_start = mocker.patch.object(Master, "start_workers")
    master = Master("my-identity.json")
    master.restart()
    mocked_stop.assert_called()
    mocked_start.assert_called()


def test_iter(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    master = Master("my-identity.json")
    assert iter(master) == master


def test_next(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    def watcher() -> Generator:
        yield {(Change.modified, "/file1.py")}
        yield None

    master = Master("my-identity.json")
    master.watcher = watcher()
    assert next(master) == [Path("/file1.py")]
    assert next(master) is None


def test_run(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    mocked_start = mocker.patch.object(Master, "start")
    mocked_stop = mocker.patch.object(Master, "stop")
    master = Master("my-identity.json")
    mocked_stop_event = mocker.MagicMock()
    master.stop_event = mocked_stop_event
    master.run()
    mocked_start.assert_called_once()
    mocked_stop_event.wait.assert_called_once()
    mocked_stop.assert_called_once()


def test_run_with_reload(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    mocker.patch.object(Master, "start")
    mocker.patch.object(Master, "stop")
    mocked_restart = mocker.patch.object(Master, "restart")

    def watcher() -> Generator:
        yield {(Change.modified, "/file1.py")}
        yield None

    master = Master("my-identity.json", server_reload=True)
    master.watcher = watcher()
    master.run()

    mocked_restart.assert_called_once()


def test_monitor_workers_restarts_dead_process(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    mocker.patch("mrok.proxy.master.MONITOR_THREAD_CHECK_DELAY", 0.1)

    mock_start_worker = mocker.patch.object(Master, "start_worker")
    master = Master("my-identity.json")

    dead_process = mocker.Mock()
    dead_process.is_alive.return_value = False
    dead_process.pid = 12345

    alive_process = mocker.Mock()
    alive_process.is_alive.return_value = True
    alive_process.pid = 12346

    new_process = mocker.Mock()
    new_process.pid = 12347

    master.worker_processes = {"id1": dead_process, "id2": alive_process}
    master.pause_event.set()

    mock_start_worker.return_value = new_process

    monitor_thread = threading.Thread(target=master.monitor_workers)
    monitor_thread.start()
    time.sleep(0.1)
    master.stop_event.set()
    monitor_thread.join()

    dead_process.stop.assert_called_once_with(sigint_timeout=1, sigkill_timeout=1)
    mock_start_worker.assert_called_once_with("id1")
    assert master.worker_processes["id1"] == new_process
    assert master.worker_processes["id2"] == alive_process


def test_monitor_workers_handles_is_alive_exception(mocker: MockerFixture):
    class Master(MasterBase):
        def get_asgi_app(self):
            return mocker.AsyncMock()

    mocker.patch("mrok.proxy.master.MONITOR_THREAD_ERROR_DELAY", 0.1)
    mock_logger = mocker.patch("mrok.proxy.master.logger.error")
    master = Master("my-identity.json")

    problematic_process = mocker.Mock()
    problematic_process.is_alive.side_effect = Exception("Test exception")

    master.worker_processes = {"id": problematic_process}
    master.pause_event.set()

    monitor_thread = threading.Thread(target=master.monitor_workers)
    monitor_thread.start()

    time.sleep(0.1)
    master.stop_event.set()
    monitor_thread.join()

    assert mock_logger.mock_calls[0].args[0] == "Error in worker monitoring: Test exception"



from pytest_mock import MockerFixture
from uvicorn.protocols.http.httptools_impl import HttpToolsProtocol as UvHttpToolsProtocol

from mrok.proxy.ziticorn import HttpToolsProtocol


def test_protocol(mocker: MockerFixture):
    mocked_super_init = mocker.patch.object(UvHttpToolsProtocol, "__init__")
    proto = HttpToolsProtocol("10", test=True)
    mocked_super_init.assert_called_once_with("10", test=True)
    assert proto.access_logger.name == "mrok.access"
    assert proto.logger.name == "mrok.proxy"
    assert proto.access_log == proto.access_logger.hasHandlers()



import logging
from pathlib import Path

from mrok.agent.sidecar.app import SidecarProxyApp
from mrok.proxy.master import MasterBase

logger = logging.getLogger("mrok.proxy")


class SidecarAgent(MasterBase):
    def __init__(
        self,
        identity_file: str,
        target: str | Path | tuple[str, int],
        ziti_load_timeout_ms: int = 5000,
        server_workers: int = 4,
        server_backlog: int = 2048,
        server_timeout_keep_alive: int = 5,
        server_limit_concurrency: int | None = None,
        server_limit_max_requests: int | None = None,
        events_enabled: bool = True,
        events_publishers_port: int = 50000,
        events_subscribers_port: int = 50001,
        events_metrics_collect_interval: float = 5.0,
        upstream_max_connections: int | None = 10,
        upstream_max_keepalive_connections: int | None = None,
        upstream_keepalive_expiry: float | None = None,
        upstream_max_connect_retries: int = 0,
    ):
        super().__init__(
            identity_file,
            ziti_load_timeout_ms=ziti_load_timeout_ms,
            server_workers=server_workers,
            server_reload=False,
            server_backlog=server_backlog,
            server_timeout_keep_alive=server_timeout_keep_alive,
            server_limit_concurrency=server_limit_concurrency,
            server_limit_max_requests=server_limit_max_requests,
            events_enabled=events_enabled,
            events_pub_port=events_publishers_port,
            events_sub_port=events_subscribers_port,
            events_metrics_collect_interval=events_metrics_collect_interval,
        )
        self._target = target
        self._max_connections = upstream_max_connections
        self._max_keepalive_connections = upstream_max_keepalive_connections
        self._keepalive_expiry = upstream_keepalive_expiry
        self._retries = upstream_max_connect_retries

    def get_asgi_app(self):
        return SidecarProxyApp(
            self._target,
            max_connections=self._max_connections,
            max_keepalive_connections=self._max_keepalive_connections,
            keepalive_expiry=self._keepalive_expiry,
            retries=self._retries,
        )


def run(
    identity_file: str,
    target_addr: str | Path | tuple[str, int],
    ziti_load_timeout_ms: int = 5000,
    server_workers: int = 4,
    server_backlog: int = 2048,
    server_timeout_keep_alive: int = 5,
    server_limit_concurrency: int | None = None,
    server_limit_max_requests: int | None = None,
    events_enabled: bool = True,
    events_publishers_port: int = 50000,
    events_subscribers_port: int = 50001,
    events_metrics_collect_interval: float = 5.0,
    upstream_max_connections: int | None = 10,
    upstream_max_keepalive_connections: int | None = None,
    upstream_keepalive_expiry: float | None = None,
    upstream_max_connect_retries: int = 0,
):
    agent = SidecarAgent(
        identity_file,
        target_addr,
        ziti_load_timeout_ms=ziti_load_timeout_ms,
        server_workers=server_workers,
        server_backlog=server_backlog,
        server_timeout_keep_alive=server_timeout_keep_alive,
        server_limit_concurrency=server_limit_concurrency,
        server_limit_max_requests=server_limit_max_requests,
        events_enabled=events_enabled,
        events_publishers_port=events_publishers_port,
        events_subscribers_port=events_subscribers_port,
        events_metrics_collect_interval=events_metrics_collect_interval,
        upstream_max_connections=upstream_max_connections,
        upstream_max_keepalive_connections=upstream_max_keepalive_connections,
        upstream_keepalive_expiry=upstream_keepalive_expiry,
        upstream_max_connect_retries=upstream_max_connect_retries,
    )
    agent.run()



import base64
import secrets
from typing import Any

import pytest

from mrok.proxy.models import FixedSizeByteBuffer, HTTPHeaders, HTTPRequest, Identity


def test_identity(ziti_identity_json: dict[str, Any], ziti_identity_file: str):
    identity = Identity.load_from_file(ziti_identity_file)
    assert identity.zt_api == ziti_identity_json["ztAPI"]
    assert identity.id.key == ziti_identity_json["id"]["key"][4:]
    assert identity.id.cert == ziti_identity_json["id"]["cert"][4:]
    assert identity.id.ca == ziti_identity_json["id"]["ca"][4:]
    assert identity.mrok.extension == ziti_identity_json["mrok"]["extension"]
    assert identity.mrok.instance == ziti_identity_json["mrok"]["instance"]


def test_fixed_size_buffer():
    chunks = [secrets.token_bytes(10) for _ in range(10)]
    fsb = FixedSizeByteBuffer(100)
    for idx in range(10):
        fsb.write(chunks[idx])

    assert fsb.getvalue() == b"".join(chunks)
    assert fsb.overflow is False


def test_fixed_size_buffer_write_no_data():
    fsb = FixedSizeByteBuffer(100)
    fsb.write(b"")
    assert fsb.getvalue() == b""
    fsb.write(None)
    assert fsb.getvalue() == b""


def test_fixed_size_buffer_overflow():
    chunks = [secrets.token_bytes(7) for _ in range(15)]
    fsb = FixedSizeByteBuffer(100)
    for idx in range(15):
        fsb.write(chunks[idx])

    assert len(fsb.getvalue()) == 100
    assert fsb.overflow is True


def test_fixed_size_buffer_full():
    chunks = [secrets.token_bytes(10) for _ in range(10)]
    fsb = FixedSizeByteBuffer(100)
    for idx in range(10):
        fsb.write(chunks[idx])

    assert fsb.getvalue() == b"".join(chunks)
    assert fsb.overflow is False
    fsb.write(secrets.token_bytes(10))
    assert fsb.getvalue() == b"".join(chunks)
    assert fsb.overflow is True


def test_fixed_size_buffer_clear():
    chunks = [secrets.token_bytes(10) for _ in range(10)]
    fsb = FixedSizeByteBuffer(100)
    for idx in range(10):
        fsb.write(chunks[idx])

    assert len(fsb.getvalue()) == 100
    assert fsb.overflow is False
    fsb.write(secrets.token_bytes(10))
    assert fsb.getvalue() == b"".join(chunks)
    assert fsb.overflow is True
    fsb.clear()
    assert fsb.getvalue() == b""
    assert fsb.overflow is False


def test_http_headers():
    headers = HTTPHeaders()
    headers["content-type"] = "application/json"

    assert headers["Content-tYpe"] == "application/json"
    assert headers.get("CONTENT-TYPE", "boh") == "application/json"
    assert headers.get("Accept", "whatever") == "whatever"
    assert headers.get("Accept") is None
    del headers["Content-tYpe"]
    with pytest.raises(KeyError):
        headers["content-type"]

    headers = HTTPHeaders(
        initial={"x-initial-header": "my-value"},
    )
    assert headers["X-Initial-Header"] == "my-value"


def test_http_headers_from_asgi():
    headers = HTTPHeaders.from_asgi(
        [
            (b"content-type", b"text/html"),
            (b"aCCept", b"application/json"),
        ]
    )

    assert headers["Content-Type"] == "text/html"
    assert headers["Accept"] == "application/json"


def test_httprequest_serialization():
    headers = HTTPHeaders.from_asgi(
        [
            (b"content-type", b"text/html"),
            (b"aCCept", b"application/json"),
        ]
    )

    req = HTTPRequest(
        method="GET",
        url="/path/to/resource",
        headers=headers,
        query_string=b"foo=bar",
        start_time=2.0,
        body=b"This is the body",
        body_truncated=False,
    )

    serialized = req.model_dump()
    assert serialized == {
        "method": "GET",
        "url": "/path/to/resource",
        "headers": {"accept": "application/json", "content-type": "text/html"},
        "query_string": base64.b64encode(b"foo=bar").decode(),
        "body": base64.b64encode(
            b"This is the body",
        ).decode(),
        "start_time": 2.0,
        "body_truncated": False,
    }


def test_httprequest_deserialization():
    serialized = {
        "method": "GET",
        "url": "/path/to/resource",
        "headers": {"accept": "application/json", "content-type": "text/html"},
        "query_string": base64.b64encode(b"foo=bar").decode(),
        "body": base64.b64encode(b"This is the body").decode(),
        "start_time": 2.0,
        "body_truncated": False,
    }
    req = HTTPRequest.model_validate(serialized)

    assert req.method == "GET"
    assert req.url == "/path/to/resource"
    assert len(req.headers) == 2
    assert req.headers["accept"] == "application/json"
    assert req.headers["content-type"] == "text/html"
    assert req.query_string == b"foo=bar"
    assert req.body == b"This is the body"
    assert req.start_time == 2.0
    assert req.body_truncated is False



from abc import ABC, abstractmethod
from typing import Any

from dynaconf.utils.boxing import DynaBox
from fastapi import HTTPException, Request, status
from fastapi.security import HTTPAuthorizationCredentials
from fastapi.security.http import HTTPBase
from pydantic import BaseModel

UNAUTHORIZED_EXCEPTION = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED, detail="Unauthorized."
)


class AuthIdentity(BaseModel):
    subject: str
    scopes: list[str] = []
    metadata: dict[str, Any] = {}


class BaseHTTPAuthBackend(ABC):
    def __init__(self, config: DynaBox):
        self.config = config
        self.scheme = self.init_scheme()

    @abstractmethod
    def init_scheme(self) -> HTTPBase:
        raise NotImplementedError()

    @abstractmethod
    async def authenticate(self, credentials: HTTPAuthorizationCredentials) -> AuthIdentity | None:
        raise NotImplementedError()

    async def __call__(self, request: Request) -> AuthIdentity | None:
        credentials = await self.scheme(request)
        if not credentials:
            return None
        return await self.authenticate(credentials)



import asyncio
from collections.abc import Iterable
from pathlib import Path

import openziti
from httpcore import SOCKET_OPTION, AsyncNetworkBackend, AsyncNetworkStream
from openziti.context import ZitiContext

from mrok.proxy.exceptions import InvalidTargetError, TargetUnavailableError
from mrok.proxy.stream import AIONetworkStream


class AIOZitiNetworkBackend(AsyncNetworkBackend):
    def __init__(self, identity_file: str | Path) -> None:
        self._identity_file = identity_file
        self._ziti_ctx: ZitiContext | None = None

    def _get_ziti_ctx(self) -> ZitiContext:
        if self._ziti_ctx is None:
            ctx, err = openziti.load(str(self._identity_file), timeout=10_000)
            if err != 0:
                raise Exception(f"Cannot create a Ziti context from the identity file: {err}")
            self._ziti_ctx = ctx
        return self._ziti_ctx

    async def connect_tcp(
        self,
        host: str,
        port: int,
        timeout: float | None = None,
        local_address: str | None = None,
        socket_options: Iterable[SOCKET_OPTION] | None = None,
    ) -> AsyncNetworkStream:
        ctx = self._get_ziti_ctx()
        try:
            sock = ctx.connect(host)
            reader, writer = await asyncio.open_connection(sock=sock)
            return AIONetworkStream(reader, writer)
        except Exception as e:
            if e.args and e.args[0] == -24:  # the service exists but is not available
                raise TargetUnavailableError() from e
            raise InvalidTargetError() from e

    async def sleep(self, seconds: float) -> None:
        await asyncio.sleep(seconds)



from pathlib import Path
from typing import Annotated

import typer

from mrok.agent import ziticorn
from mrok.cli.utils import number_of_workers

default_workers = number_of_workers()


def register(app: typer.Typer) -> None:
    @app.command("asgi")
    def run_asgi(
        identity_file: Annotated[Path, typer.Argument(..., help="Identity json file")],
        app: Annotated[str, typer.Argument(..., help="ASGI application")],
        ziti_load_timeout_ms: Annotated[
            int,
            typer.Option(
                "--ziti-load-timeout-ms",
                help="Timeout (ms) waiting for Ziti to load.",
                show_default=True,
            ),
        ] = 5000,
        server_workers: Annotated[
            int,
            typer.Option(
                "--server-workers",
                "-w",
                help=f"Number of workers. Default: {default_workers}",
                show_default=True,
            ),
        ] = default_workers,
        server_backlog: Annotated[
            int,
            typer.Option(
                "--server-backlog",
                help="TCP socket listen backlog.",
                show_default=True,
            ),
        ] = 2048,
        server_timeout_keep_alive: Annotated[
            int,
            typer.Option(
                "--server-timeout-keep-alive",
                help="Seconds to keep idle HTTP connections open.",
                show_default=True,
            ),
        ] = 5,
        server_limit_concurrency: Annotated[
            int | None,
            typer.Option(
                "--server-limit-concurrency",
                help="Maximum number of concurrent requests per worker.",
                show_default=True,
            ),
        ] = None,
        server_limit_max_requests: Annotated[
            int | None,
            typer.Option(
                "--server-limit-max-requests",
                help="Restart a worker after handling this many requests.",
                show_default=True,
            ),
        ] = None,
        server_reload: Annotated[
            bool,
            typer.Option(
                "--server-reload",
                "-r",
                help="Enable server auto-reload. Default: False",
                show_default=True,
            ),
        ] = False,
        events_publishers_port: Annotated[
            int,
            typer.Option(
                "--events-publishers-port",
                help=(
                    "TCP port where the mrok agent "
                    "should connect to publish to request/response messages."
                ),
                show_default=True,
            ),
        ] = 50000,
        events_subscribers_port: Annotated[
            int,
            typer.Option(
                "--events-subscribers-port",
                help=(
                    "TCP port where the mrok agent should listen for incoming subscribers "
                    "connections for request/response messages."
                ),
                show_default=True,
            ),
        ] = 50001,
        events_metrics_collect_interval: Annotated[
            float,
            typer.Option(
                "--events-metrics-collect-interval",
                help="Interval in seconds between events metrics collect.",
                show_default=True,
            ),
        ] = 5.0,
    ):
        """Run an ASGI application exposing it through OpenZiti network."""
        ziticorn.run(
            app,
            str(identity_file),
            ziti_load_timeout_ms=ziti_load_timeout_ms,
            server_workers=server_workers,
            server_reload=server_reload,
            server_backlog=server_backlog,
            server_timeout_keep_alive=server_timeout_keep_alive,
            server_limit_concurrency=server_limit_concurrency,
            server_limit_max_requests=server_limit_max_requests,
            events_metrics_collect_interval=events_metrics_collect_interval,
            events_publishers_port=events_publishers_port,
            events_subscribers_port=events_subscribers_port,
        )



import pytest
from pytest_mock import MockerFixture

from mrok.proxy.ziticorn import BackendConfig, Server


@pytest.mark.asyncio
async def test_serve(mocker: MockerFixture, ziti_identity_file: str):
    mocked_socket = mocker.MagicMock()
    mocker.patch.object(BackendConfig, "bind_socket", return_value=mocked_socket)
    mocked_inner_serve = mocker.patch.object(Server, "_serve")

    async def fake_asgi_app(scope, receive, send):
        pass

    config = BackendConfig(fake_asgi_app, ziti_identity_file)

    server = Server(config)
    await server.serve()
    mocked_inner_serve.assert_awaited_once_with([mocked_socket])


@pytest.mark.asyncio
async def test_serve_with_socket(mocker: MockerFixture, ziti_identity_file: str):
    mocked_socket = mocker.MagicMock()
    mocked_bind = mocker.patch.object(BackendConfig, "bind_socket")
    mocked_inner_serve = mocker.patch.object(Server, "_serve")

    async def fake_asgi_app(scope, receive, send):
        pass

    config = BackendConfig(fake_asgi_app, ziti_identity_file)

    server = Server(config)
    await server.serve([mocked_socket])
    mocked_inner_serve.assert_awaited_once_with([mocked_socket])
    mocked_bind.assert_not_called()



from contextlib import asynccontextmanager

import pytest
from pytest_mock import MockerFixture

from mrok.proxy.asgi import ASGIAppWrapper
from mrok.types.proxy import Message
from tests.types import ReceiveFactory, SendFactory


def test_wrapper_middleware_stack(mocker: MockerFixture):
    m_app = mocker.MagicMock()

    m_mw1 = mocker.MagicMock()
    m_mw1_ctor = mocker.MagicMock(return_value=m_mw1)
    args1 = ("a", True)
    kwargs1 = {"mykwargs1": "value1"}

    m_mw2 = mocker.MagicMock()
    m_mw2_ctor = mocker.MagicMock(return_value=m_mw2)
    args2 = ("b", False)
    kwargs2 = {"mykwargs2": "value2"}

    wrapper = ASGIAppWrapper(m_app)
    wrapper.add_middleware(m_mw1_ctor, *args1, **kwargs1)  # type: ignore
    wrapper.add_middleware(m_mw2_ctor, *args2, **kwargs2)  # type: ignore

    stack = wrapper.build_middleware_stack()
    assert stack == m_mw2

    m_mw1_ctor.assert_called_once_with(m_app, *args1, **kwargs1)
    m_mw2_ctor.assert_called_once_with(m_mw1, *args2, **kwargs2)


async def test_wrapper_lifespan_stack(
    mocker: MockerFixture,
    receive_factory: ReceiveFactory,
    send_factory: SendFactory,
):
    m_inner_startup = mocker.AsyncMock()
    m_inner_shutdown = mocker.AsyncMock()

    @asynccontextmanager
    async def inner_lifespan(app):
        await m_inner_startup()
        yield
        await m_inner_shutdown()

    m_outer_startup = mocker.AsyncMock()
    m_outer_shutdown = mocker.AsyncMock()

    @asynccontextmanager
    async def outer_lifespan(app):
        await m_outer_startup()
        yield {"outer_state": "a state"}
        await m_outer_shutdown()

    m_wrapper_router = mocker.AsyncMock(lifespan_context=inner_lifespan)
    m_wrapped_app = mocker.AsyncMock(router=m_wrapper_router)

    sent: list[Message] = []
    receive = receive_factory([{"type": "lifespan.startup"}, {"type": "lifespan.shutdown"}])
    send = send_factory(sent)

    wrapper = ASGIAppWrapper(
        m_wrapped_app,
        lifespan=outer_lifespan,
    )

    await wrapper({"type": "lifespan", "state": {}}, receive, send)
    assert sent == [
        {"type": "lifespan.startup.complete"},
        {"type": "lifespan.shutdown.complete"},
    ]
    m_inner_startup.assert_awaited_once()
    m_inner_shutdown.assert_awaited_once()
    m_outer_startup.assert_awaited_once()
    m_outer_shutdown.assert_awaited_once()


async def test_wrapper_no_app_lifespan(
    mocker: MockerFixture,
    receive_factory: ReceiveFactory,
    send_factory: SendFactory,
):
    m_outer_startup = mocker.AsyncMock()
    m_outer_shutdown = mocker.AsyncMock()

    @asynccontextmanager
    async def outer_lifespan(app):
        await m_outer_startup()
        yield {"outer_state": "a state"}
        await m_outer_shutdown()

    m_wrapper_router = mocker.AsyncMock(lifespan_context=None)
    m_wrapped_app = mocker.AsyncMock(router=m_wrapper_router)

    sent: list[Message] = []
    receive = receive_factory([{"type": "lifespan.startup"}, {"type": "lifespan.shutdown"}])
    send = send_factory(sent)

    wrapper = ASGIAppWrapper(
        m_wrapped_app,
        lifespan=outer_lifespan,
    )

    await wrapper({"type": "lifespan", "state": {}}, receive, send)
    assert sent == [
        {"type": "lifespan.startup.complete"},
        {"type": "lifespan.shutdown.complete"},
    ]
    m_outer_startup.assert_awaited_once()
    m_outer_shutdown.assert_awaited_once()


async def test_wrapper_no_starlette_app(
    mocker: MockerFixture,
    receive_factory: ReceiveFactory,
    send_factory: SendFactory,
):
    m_outer_startup = mocker.AsyncMock()
    m_outer_shutdown = mocker.AsyncMock()

    @asynccontextmanager
    async def outer_lifespan(app):
        await m_outer_startup()
        yield {"outer_state": "a state"}
        await m_outer_shutdown()

    m_wrapped_app = mocker.AsyncMock(router=None)

    sent: list[Message] = []
    receive = receive_factory([{"type": "lifespan.startup"}, {"type": "lifespan.shutdown"}])
    send = send_factory(sent)

    wrapper = ASGIAppWrapper(
        m_wrapped_app,
        lifespan=outer_lifespan,
    )

    await wrapper({"type": "lifespan", "state": {}}, receive, send)
    assert sent == [
        {"type": "lifespan.startup.complete"},
        {"type": "lifespan.shutdown.complete"},
    ]
    m_outer_startup.assert_awaited_once()
    m_outer_shutdown.assert_awaited_once()


@pytest.mark.asyncio
async def test_wrapper_no_lifespan(
    mocker: MockerFixture,
):
    m_app = mocker.AsyncMock(router=None)

    m_receive = mocker.AsyncMock()
    m_send = mocker.AsyncMock()

    wrapper = ASGIAppWrapper(m_app)

    await wrapper({"type": "lifespan"}, m_receive, m_send)
    m_app.assert_not_awaited()


@pytest.mark.asyncio
async def test_wrapper_invoke_app(
    mocker: MockerFixture,
    receive_factory: ReceiveFactory,
):
    m_app = mocker.AsyncMock()

    receive = receive_factory()
    m_send = mocker.AsyncMock()

    wrapper = ASGIAppWrapper(m_app)

    await wrapper({"type": "http"}, receive, m_send)
    m_app.assert_awaited_once_with({"type": "http"}, receive, m_send)


async def test_wrapper_lifespan_state_unsupported(
    mocker: MockerFixture,
    receive_factory: ReceiveFactory,
    send_factory: SendFactory,
):
    m_outer_startup = mocker.AsyncMock()
    m_outer_shutdown = mocker.AsyncMock()

    @asynccontextmanager
    async def outer_lifespan(app):
        await m_outer_startup()
        yield {"outer_state": "a state"}
        await m_outer_shutdown()

    m_wrapped_app = mocker.AsyncMock(router=None)

    sent: list[Message] = []
    receive = receive_factory([{"type": "lifespan.startup"}, {"type": "lifespan.shutdown"}])
    send = send_factory(sent)

    wrapper = ASGIAppWrapper(
        m_wrapped_app,
        lifespan=outer_lifespan,
    )
    with pytest.raises(Exception) as cv:
        await wrapper({"type": "lifespan"}, receive, send)

    assert str(cv.value) == '"state" is unsupported by the current ASGI Server.'



from mrok.controller.auth.base import BaseHTTPAuthBackend

BACKEND_REGISTRY: dict[str, type[BaseHTTPAuthBackend]] = {}


def register_authentication_backend(name: str):
    """Decorator to register a backend class with a unique key."""

    def decorator(cls: type[BaseHTTPAuthBackend]):
        BACKEND_REGISTRY[name] = cls
        return cls

    return decorator


def get_authentication_backend(name: str) -> type[BaseHTTPAuthBackend] | None:
    return BACKEND_REGISTRY.get(name)



import asyncio
import base64
import json
import tempfile
from collections.abc import AsyncGenerator, Generator
from datetime import UTC, datetime, timedelta
from typing import Any

import jwt
import pytest
import zmq
from asgi_lifespan import LifespanManager
from dynaconf import Dynaconf
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from pytest_httpx import HTTPXMock

from mrok.conf import Settings, get_settings
from mrok.types.proxy import ASGIReceive, ASGISend, Message
from tests.types import ReceiveFactory, SendFactory, SettingsFactory, StatusEventFactory


@pytest.fixture(scope="session")
def settings_factory() -> SettingsFactory:
    def _get_settings(
        ziti: dict | None = None,
        logging: dict | None = None,
        frontend: dict | None = None,
        controller: dict | None = None,
    ) -> Settings:
        ziti = ziti or {
            "base_urls": {
                "management": "https://ziti.example.com",
                "client": "https://ziti.example.com",
            },
            "connect_timeout": 0.25,
            "read_timeout": 10,
            "ssl_verify": True,
            "auth": {"username": "user", "password": "pass", "identity": None},
        }
        logging = logging or {
            "debug": True,
            "rich": False,
        }

        controller = controller or {}
        if "auth" not in controller:
            controller["auth"] = {
                "backends": ["oidc"],
                "oidc": {
                    "config_url": "http://example.com/openid-configuration",
                    "audience": "mrok-audience",
                    "subject_claim": "sub",
                },
            }
        if "pagination" not in controller:
            controller["pagination"] = {"limit": 5}
        frontend = frontend or {
            "identity": "public",
            "mode": "zrok",
            "domain": "exts.s1.today",
        }
        settings = Dynaconf(
            environments=True,
            settings_files=[],
            ENV_FOR_DYNACONF="testing",
            ZITI=ziti,
            LOGGING=logging,
            FRONTEND=frontend,
            CONTROLLER=controller,
        )

        return settings

    return _get_settings


@pytest.fixture()
def ziti_bad_request_error() -> dict[str, Any]:
    return {
        "error": {
            "code": "error_code",
            "message": "error_message",
            "cause": {
                "field": "field_name",
                "reason": "field_error",
            },
        }
    }


@pytest.fixture()
def ziti_identity_json() -> dict[str, Any]:
    return {
        "ztAPI": "https://ziti.platform.softwareone.com/edge/client/v1",
        "ztAPIs": None,
        "configTypes": None,
        "id": {
            "key": "pem:-----BEGIN PRIVATE KEY-----\n...\n-----END PRIVATE KEY-----\n",
            "cert": "pem:-----BEGIN CERTIFICATE-----\n...\n-----END CERTIFICATE-----\n",
            "ca": "pem:-----BEGIN CERTIFICATE-----\n...\n-----END CERTIFICATE-----\n",
        },
        "enableHa": False,
        "mrok": {
            "extension": "ext-0000-0000",
            "instance": "ins-0000-0000-0000",
            "domain": "exts.platform.softwareone.com",
            "tags": {
                "mrok-service": "ext-0000-0000",
                "mrok-identity-type": "instance",
                "mrok": "0.4.0",
            },
        },
    }


@pytest.fixture()
def ziti_identity_file(ziti_identity_json: dict[str, Any]) -> Generator[str, None, None]:
    with tempfile.NamedTemporaryFile("w", suffix="json") as f:
        json.dump(ziti_identity_json, f)
        f.seek(0)
        yield f.name


@pytest.fixture()
def jwt_signing_key() -> str:
    return """-----BEGIN PRIVATE KEY-----
MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQDh0IIuMVuxXSq1
ndmO13RLpqv+3Ndyh0aEP8Lz3/zQeyliYCfW8HMoP8vjC9ftGKdEHLG4eoUs6NF+
8dYh3MRtDikl0oH/J+/6DFS3zqG+NNaMUpTgmKsrn1w5MN/NPQxPp/Djc4nAcAoi
V/c0w5fft5CLoLkLt5iL8y2mTuR0VRhZ6fS+hQiQMcSBajtr8jPLhem9spJ53N5x
X/OQNTkP712s9qSwCLd8SoFzNFYam2kHKs+nfMbXmk0PXbLjTb9Sk6pOa4Ey5OZU
//8oFyMnEltuRUH158WskpxSsxwZosPOsrFXcEdq5DHPQppc5H6IjwevVr8VZCZ9
y0DhqbpHAgMBAAECggEAa23BiQ1VVIeA2p9qkbDBtH3qJJlW7DccHq2g51nf0iVN
1m3tdi21c2gKbQ1E4BjS9q36BMxREEEA7bZKy5hWYJWUkNhZnRSYb+qu4TUuMKm9
ikt9ZW5sPJgXjWPJjUAmW70HdoYQelszDwyvYExPNBEF9M91SXRCYA5EYPL2b7rp
FjUzbq4/a4C9GYUP7EhBYYyZ+KnFDa9mxdbUFrQgupV2b+v2sPeOmkDRhzjAp7HC
yRcUnTmQM32LMO57oFZ7K9DJDMwxbHCAE7N8f2Uge2VLpIUjz3GskdjPcLBqu4qH
938dL3GMvKglo18VRSKiTkwxjxnP8HA8+0ifRv7cIQKBgQD7irdN/lYXdlhKAXwt
eQ5Tua3/Zi3Xia0dgOEzGxfVT1L6rVjVa149/v7x26p0evt0c9gkcHlBgUDEH7yz
sEDY/KwIo9Wg2Bpe1iUNcO/BvZc0mc3Cu9D8agxNZ5eHq+a36gTKR/QD1dp+48zK
6yimXxEfUYy2u8TBLm2Jk3hx4QKBgQDl0RA+L/SP8d2lB4vPDSSFhMYnwaH4njbq
/S8evNFcyKba9HaA8Sbng/msI/k8rnyul5GqFVooVl9TfPa93ATvagzijBvvqZ6R
j2+41hEAxmZ2257DKYpQ4JM2mRl9pOFQrc1GRbUfSOJAekj2O/2ojnbHHUcxg0dS
yliskq6BJwKBgGxVDKcBb5CBPnr48sMezMXQRRimp/2Y5L69H8AD3hrXI/SkLYsU
x6zJooEFSv8JbDx2G9NtwTst8HfG910n/nW1NF4wOTQhfhH0BlcomYmGHpXf25cP
jmz3Oz8m60LaDO6OUevQW04/ju9xKmUGLCai8NvdIk4cxhsw5KoIoinhAoGBALDu
vCqkkQ0hkRs1LBZEcBG7nzOMiD740B8qvdRUWnusn4mDHJk5EFK98MLvDzwAuk1Q
s/zWY4satFl6pByX/9SzOShR5lAlrscyPzl21bBbDxgDDcADg1GxFKW8STvKbQ3I
QXoQwNlNK6OogfPRTAExbZDuoZklEQxUbOCwLVmRAoGACIi7DWWZr9Ea4ag82yYv
CBdlKSshb2RYGdbYm9dzeqqv1/xtwtFYMOuLilL3qp9erBAsc6Kdz+n3v1+2p7Ho
3/upcFI5QbhGHCO5J92fdDYh6rsAG4cwC5LlAC5WgP3bL9+QVKKcYm8/mmglLIu3
XIvgj23tDAmXtVRh0MsgOH4=
-----END PRIVATE KEY-----
"""


@pytest.fixture()
def jwt_token(jwt_signing_key: str) -> str:
    payload = {
        "sub": "user123",
        "aud": "mrok-audience",
        "iss": "http:/test.mrok/",
        "exp": datetime.now(UTC) + timedelta(minutes=30),
    }
    return jwt.encode(payload, jwt_signing_key, algorithm="RS256", headers={"kid": "test-key"})


@pytest.fixture()
def jwks_json() -> dict:
    return {
        "keys": [
            {
                "kty": "RSA",
                "use": "sig",
                "alg": "RS256",
                "kid": "test-key",
                "n": (
                    "4dCCLjFbsV0qtZ3Zjtd0S6ar_tzXcodGhD_C89_80HspYmAn1vBzKD_"
                    "L4wvX7RinRByxuHqFLOjRfvHWIdzEbQ4pJdKB_yfv-gxUt86hvjTWjF"
                    "KU4JirK59cOTDfzT0MT6fw43OJwHAKIlf3NMOX37eQi6C5C7eYi_Mtp"
                    "k7kdFUYWen0voUIkDHEgWo7a_Izy4XpvbKSedzecV_zkDU5D-9drPak"
                    "sAi3fEqBczRWGptpByrPp3zG15pND12y402_UpOqTmuBMuTmVP__KBc"
                    "jJxJbbkVB9efFrJKcUrMcGaLDzrKxV3BHauQxz0KaXOR-iI8Hr1a_FW"
                    "QmfctA4am6Rw"
                ),
                "e": "AQAB",
            }
        ]
    }


@pytest.fixture()
def openid_config() -> dict:
    return {"issuer": "http:/test.mrok/", "jwks_uri": "http://example.com/jwks.json"}


@pytest.fixture()
def fastapi_app(settings_factory: SettingsFactory) -> FastAPI:
    settings = settings_factory()
    from mrok.controller.app import setup_app

    app = setup_app(settings)
    app.dependency_overrides[get_settings] = lambda: settings
    return app


@pytest.fixture()
async def app_lifespan_manager(fastapi_app: FastAPI) -> AsyncGenerator[LifespanManager, None]:
    async with LifespanManager(fastapi_app) as lifespan_manager:
        yield lifespan_manager


@pytest.fixture
async def api_client(
    fastapi_app: FastAPI,
    app_lifespan_manager: LifespanManager,
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
    openid_config: dict,
    jwks_json: dict,
    jwt_token: str,
) -> AsyncGenerator[AsyncClient]:
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=settings.controller.auth.oidc.config_url,
        json=openid_config,
        is_reusable=True,
    )
    httpx_mock.add_response(
        method="GET",
        url="http://example.com/jwks.json",
        json=jwks_json,
        is_reusable=True,
    )

    async with AsyncClient(
        transport=ASGITransport(app=app_lifespan_manager.app),
        base_url=f"http://localhost/{fastapi_app.root_path.removeprefix('/')}/",
        headers={"Authorization": f"Bearer {jwt_token}"},
    ) as client:
        yield client


@pytest.fixture
def receive_factory() -> ReceiveFactory:
    def _factory(messages: list[Message] | None = None) -> ASGIReceive:
        if not messages:
            messages = [{"type": "http.request", "body": b"", "more_body": False}]

        class Receiver:
            def __init__(self, messages: list[Message]):
                self.messages = messages

            async def __call__(self):
                await asyncio.sleep(0)
                try:
                    return self.messages.pop(0)
                except Exception:
                    return

        return Receiver(messages)

    return _factory


@pytest.fixture
def send_factory() -> SendFactory:
    def _factory(collected: list[Message]) -> ASGISend:
        async def send(message: Message) -> None:
            await asyncio.sleep(0)
            collected.append(message)

        return send

    return _factory


@pytest.fixture()
def response_event_factory():
    def _response_event(
        method: str = "GET",
        url: str = "/test",
        request_headers: dict[str, str] | None = None,
        request_querystring: bytes | None = None,
        request_body: bytes | None = None,
        request_truncated: bool = False,
        response_headers: dict[str, str] | None = None,
        response_status: int = 200,
        response_body: bytes | None = b'{"test": "json"}',
        response_truncated: bool = False,
        duration: float = 10.3,
    ) -> dict[str, Any]:
        request_headers = request_headers or {"accept": "application/json"}
        response_headers = response_headers or {"content-type": "application/json"}
        request = {
            "method": method,
            "url": url,
            "headers": request_headers,
            "query_string": None,
            "start_time": 0.0,
            "body": None,
            "body_truncated": request_truncated,
        }

        response = {
            "type": "response",
            "data": {
                "type": "response",
                "request": request,
                "status": response_status,
                "headers": response_headers,
                "duration": duration,
                "body": None,
                "body_truncated": response_truncated,
            },
        }
        if request_querystring:
            request["query_string"] = base64.b64encode(request_querystring).decode("ascii")
        if request_body:
            request["body"] = base64.b64encode(request_body).decode("ascii")
        if response_body:
            response["data"]["body"] = base64.b64encode(response_body).decode("ascii")  # type: ignore[index]
        return response

    return _response_event


@pytest.fixture()
def status_event_factory() -> StatusEventFactory:
    def _status_event(
        process_cpu: float = 55.1,
        process_mem: float = 21.2,
        requests_rps: int = 123,
        requests_total: int = 1000,
        requests_successful: int = 10,
        requests_failed: int = 30,
        bytes_in: int = 1000,
        bytes_out: int = 2000,
    ):
        return {
            "type": "status",
            "data": {
                "type": "status",
                "meta": {
                    "extension": "EXT-1223-4443",
                    "instance": "INS-3737-8373-7373-1113-3384",
                    "domain": "ext.mrok.test",
                    "tags": None,
                },
                "metrics": {
                    "worker_id": "my-worker-id",
                    "data_transfer": {"bytes_in": bytes_in, "bytes_out": bytes_out},
                    "requests": {
                        "rps": requests_rps,
                        "total": requests_total,
                        "successful": requests_successful,
                        "failed": requests_failed,
                    },
                    "response_time": {
                        "avg": 10.0,
                        "min": 1,
                        "max": 30,
                        "p50": 11,
                        "p90": 22,
                        "p99": 11,
                    },
                    "process": {"cpu": process_cpu, "mem": process_mem},
                },
            },
        }

    return _status_event


@pytest.fixture()
def multipart_body() -> tuple[str, bytes]:
    boundary = "----WebKitFormBoundary7MA4YWxkTrZu0gW"
    data = b"""------WebKitFormBoundary7MA4YWxkTrZu0gW\r
Content-Disposition: form-data; name="username"\r
\r
john_doe\r
------WebKitFormBoundary7MA4YWxkTrZu0gW\r
Content-Disposition: form-data; name="email"\r
\r
john@example.com\r
------WebKitFormBoundary7MA4YWxkTrZu0gW\r
Content-Disposition: form-data; name="message"\r
\r
Hello, this is a test message!\r
------WebKitFormBoundary7MA4YWxkTrZu0gW\r
Content-Disposition: form-data; name="file"; filename="test.txt"\r
Content-Type: application/octet-stream\r
\r
binary file content\r
------WebKitFormBoundary7MA4YWxkTrZu0gW--\r
"""

    return boundary, data


@pytest.fixture(scope="session")
def zmq_publisher() -> Generator[tuple[zmq.Socket, int], None, None]:
    context: zmq.Context = zmq.Context.instance()
    socket: zmq.Socket = context.socket(zmq.XPUB)

    port: int = socket.bind_to_random_port("tcp://127.0.0.1")

    yield socket, port

    socket.close()
    context.term()


@pytest.fixture()
def ziti_frontend_error_template_html() -> str:
    return "{{ status }}\n{{ body }}\n"


@pytest.fixture()
def ziti_frontend_error_template_html_file(
    ziti_frontend_error_template_html: str,
) -> Generator[str, None, None]:
    with tempfile.NamedTemporaryFile("w", suffix="html") as f:
        f.write(ziti_frontend_error_template_html)
        f.seek(0)
        yield f.name


@pytest.fixture()
def ziti_frontend_error_template_json() -> str:
    return """{
        "status": {{ status }},
        "body": "{{ body }}",
    }
"""


@pytest.fixture()
def ziti_frontend_error_template_json_file(
    ziti_frontend_error_template_json: str,
) -> Generator[str, None, None]:
    with tempfile.NamedTemporaryFile("w", suffix="json") as f:
        f.write(ziti_frontend_error_template_json)
        f.seek(0)
        yield f.name



import shlex

import pytest
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from mrok.cli import app


def test_run_asgi(
    mocker: MockerFixture,
):
    mocked_ziticorn = mocker.patch("mrok.cli.commands.agent.run.asgi.ziticorn.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split(
            "agent run asgi my:app ins-1234-5678-0001.json -w 2 --server-reload "
            "--events-publishers-port 4000 "
            "--events-subscribers-port 5000"
        ),
    )
    assert result.exit_code == 0
    mocked_ziticorn.assert_called_once_with(
        "ins-1234-5678-0001.json",
        "my:app",
        ziti_load_timeout_ms=5000,
        server_workers=2,
        server_reload=True,
        server_backlog=2048,
        server_limit_concurrency=None,
        server_limit_max_requests=None,
        server_timeout_keep_alive=5,
        events_publishers_port=4000,
        events_subscribers_port=5000,
        events_metrics_collect_interval=5.0,
    )


@pytest.mark.parametrize(
    ("target_addr", "expected_target_addr"),
    [
        ("/path/to/unix.sock", "/path/to/unix.sock"),
        ("localhost:8000", ("localhost", 8000)),
        (":8000", ("127.0.0.1", 8000)),
    ],
)
def test_run_sidecar(
    mocker: MockerFixture,
    target_addr,
    expected_target_addr,
):
    mocked_sidecar = mocker.patch("mrok.cli.commands.agent.run.sidecar.sidecar.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split(
            f"agent run sidecar ins-1234-5678-0001.json {target_addr} "
            "-w 2 --events-publishers-port 4000 --events-subscribers-port 5000 "
            "--upstream-max-connections 312 "
            "--upstream-max-keepalive-connections 11 "
            "--upstream_keepalive_expiry 3.22 "
            "--upstream-max-connect-retries 2"
        ),
    )
    assert result.exit_code == 0
    mocked_sidecar.assert_called_once_with(
        "ins-1234-5678-0001.json",
        expected_target_addr,
        events_enabled=True,
        server_workers=2,
        upstream_max_connections=312,
        upstream_max_keepalive_connections=11,
        upstream_keepalive_expiry=3.22,
        upstream_max_connect_retries=2,
        events_publishers_port=4000,
        events_subscribers_port=5000,
        events_metrics_collect_interval=5.0,
        server_backlog=2048,
        server_limit_concurrency=None,
        server_limit_max_requests=None,
        server_timeout_keep_alive=5,
        ziti_load_timeout_ms=5000,
    )



import asyncio

import typer

from mrok.cli.utils import validate_extension_id, validate_instance_id
from mrok.conf import Settings, get_settings
from mrok.ziti.api import ZitiManagementAPI
from mrok.ziti.identities import unregister_identity


async def do_unregister(settings: Settings, extension_id: str, instance_id: str):
    async with ZitiManagementAPI(settings) as api:
        await unregister_identity(settings, api, extension_id, instance_id)


def register(app: typer.Typer) -> None:
    settings = get_settings()

    @app.command("instance")
    def unregister_instance(
        ctx: typer.Context,
        extension_id: str = typer.Argument(
            ...,
            callback=validate_extension_id,
            help=f"Extension ID in the format {settings.identifiers.extension.format}",
        ),
        instance_id: str = typer.Argument(
            ...,
            callback=validate_instance_id,
            help=f"Instance ID in the format {settings.identifiers.instance.format}",
        ),
    ):
        """Register a new Extension Instance in OpenZiti (identity)."""
        asyncio.run(do_unregister(ctx.obj, extension_id, instance_id))
        print(f"🍻 [green]Instance [bold]{instance_id}[/bold] unregistered.[/green]")



import pytest

from mrok.frontend.utils import parse_accept_header


@pytest.mark.parametrize(
    ("accept_header", "expected"),
    [
        ("text/html, application/json", ["text/html", "application/json"]),
        ("text/html;q=0.8, application/json;q=0.9", ["application/json", "text/html"]),
        (
            "text/plain;q=0.1, text/html, application/json;q=0.5",
            ["text/html", "application/json", "text/plain"],
        ),
        (
            "application/json;q=0.7, text/html;q=0.7, text/plain;q=0.7",
            ["application/json", "text/html", "text/plain"],
        ),
        ("*/*;q=0.1, application/json, text/*;q=0.5", ["application/json", "text/*", "*/*"]),
        (
            "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,application/json;q=0.8",
            [
                "text/html",
                "application/xhtml+xml",
                "image/webp",
                "application/xml",
                "application/json",
            ],
        ),
        ("application/json, text/html;q=0", ["application/json", "text/html"]),
        (None, ["*/*"]),
        ("", ["*/*"]),
    ],
)
def test_parse_accept_header(accept_header: str | None, expected: list[str]):
    assert parse_accept_header(accept_header) == expected


def test_wrong_accept_string():
    assert parse_accept_header(",") == []


def test_wrong_weight():
    assert parse_accept_header(
        "text/html;q=0.8, application/json;q=0.9, application/xml;q=1a4"
    ) == ["application/json", "text/html", "application/xml"]



import re

RE_EXTENSION_ID = re.compile(r"(?i)EXT-\d{4}-\d{4}")
RE_INSTANCE_ID = re.compile(r"(?i)INS-\d{4}-\d{4}-\d{4}")


BINARY_CONTENT_TYPES = {
    "application/octet-stream",
    "application/pdf",
}

BINARY_PREFIXES = (
    "image/",
    "video/",
    "audio/",
)

TEXTUAL_CONTENT_TYPES = {
    "application/json",
    "application/xml",
    "application/javascript",
    "application/x-www-form-urlencoded",
}

TEXTUAL_PREFIXES = ("text/",)



import pytest
from pytest_mock import MockerFixture

from mrok.proxy.backend import AIOZitiNetworkBackend
from mrok.proxy.exceptions import InvalidTargetError, TargetUnavailableError


@pytest.mark.asyncio
async def test_connect_tcp(mocker: MockerFixture):
    mocked_ziti_sock = mocker.MagicMock()
    mocked_ziti_ctx = mocker.MagicMock()
    mocked_ziti_ctx.connect.return_value = mocked_ziti_sock
    mocked_ziti_load = mocker.patch(
        "mrok.proxy.backend.openziti.load", return_value=(mocked_ziti_ctx, 0)
    )
    mocked_reader = mocker.MagicMock()
    mocked_writer = mocker.MagicMock()
    mocked_aio_ns = mocker.MagicMock()
    mocked_aio_netstream_ctor = mocker.patch(
        "mrok.proxy.backend.AIONetworkStream", return_value=mocked_aio_ns
    )
    mocked_aio_openconn = mocker.patch(
        "mrok.proxy.backend.asyncio.open_connection", return_value=(mocked_reader, mocked_writer)
    )

    backend = AIOZitiNetworkBackend("my_identity_file.json")
    stream = await backend.connect_tcp("ziti-svc", 0)
    mocked_ziti_load.assert_called_once_with("my_identity_file.json", timeout=10000)
    mocked_ziti_ctx.connect.assert_called_once_with("ziti-svc")
    mocked_aio_netstream_ctor.assert_called_once_with(mocked_reader, mocked_writer)
    mocked_aio_openconn.assert_called_once_with(sock=mocked_ziti_sock)
    assert stream == mocked_aio_ns


@pytest.mark.asyncio
async def test_connect_tcp_context_created_once(mocker: MockerFixture):
    mocked_ziti_sock = mocker.MagicMock()
    mocked_ziti_ctx = mocker.MagicMock()
    mocked_ziti_ctx.connect.return_value = mocked_ziti_sock
    mocked_ziti_load = mocker.patch(
        "mrok.proxy.backend.openziti.load", return_value=(mocked_ziti_ctx, 0)
    )
    mocked_reader = mocker.MagicMock()
    mocked_writer = mocker.MagicMock()
    mocked_aio_ns = mocker.MagicMock()
    mocker.patch("mrok.proxy.backend.AIONetworkStream", return_value=mocked_aio_ns)
    mocker.patch(
        "mrok.proxy.backend.asyncio.open_connection", return_value=(mocked_reader, mocked_writer)
    )

    backend = AIOZitiNetworkBackend("my_identity_file.json")
    _ = await backend.connect_tcp("ziti-svc", 0)
    _ = await backend.connect_tcp("ziti-svc2", 0)
    assert mocked_ziti_load.call_count == 1


@pytest.mark.asyncio
async def test_connect_tcp_load_context_error(mocker: MockerFixture):
    mocker.patch("mrok.proxy.backend.openziti.load", return_value=(mocker.MagicMock(), -20))

    backend = AIOZitiNetworkBackend("my_identity_file.json")
    with pytest.raises(Exception) as cv:
        await backend.connect_tcp("ziti-svc", 0)
    assert str(cv.value) == "Cannot create a Ziti context from the identity file: -20"


@pytest.mark.asyncio
async def test_connect_tcp_target_unavailable(mocker: MockerFixture):
    mocked_ziti_ctx = mocker.MagicMock()
    mocked_ziti_ctx.connect.side_effect = Exception(-24, "service unavailable")
    mocker.patch("mrok.proxy.backend.openziti.load", return_value=(mocked_ziti_ctx, 0))

    backend = AIOZitiNetworkBackend("my_identity_file.json")
    with pytest.raises(TargetUnavailableError):
        await backend.connect_tcp("ziti-svc", 0)


@pytest.mark.asyncio
async def test_connect_tcp_target_doesnt_exist(mocker: MockerFixture):
    mocked_ziti_ctx = mocker.MagicMock()
    mocked_ziti_ctx.connect.side_effect = Exception(-18, "service doesn't exist")
    mocker.patch("mrok.proxy.backend.openziti.load", return_value=(mocked_ziti_ctx, 0))

    backend = AIOZitiNetworkBackend("my_identity_file.json")
    with pytest.raises(InvalidTargetError):
        await backend.connect_tcp("ziti-svc", 0)


@pytest.mark.asyncio
async def test_sleep(mocker: MockerFixture):
    mocked_sleep = mocker.patch("mrok.proxy.backend.asyncio.sleep")
    backend = AIOZitiNetworkBackend("my_identity_file.json")
    await backend.sleep(4)
    mocked_sleep.assert_awaited_once_with(4)



import multiprocessing
import re

import typer

from mrok.conf import get_settings


def number_of_workers() -> int:
    return (multiprocessing.cpu_count() * 2) + 1


def validate_identifier(regex_exp: str, format: str, identifier: str) -> str:
    match = re.fullmatch(regex_exp, identifier)
    if not match:
        raise typer.BadParameter(f"it must match {format}")
    return identifier


def validate_extension_id(extension_id: str) -> str:
    settings = get_settings()
    return validate_identifier(
        settings.identifiers.extension.regex, settings.identifiers.extension.format, extension_id
    )


def validate_instance_id(instance_id: str) -> str:
    settings = get_settings()
    return validate_identifier(
        settings.identifiers.instance.regex, settings.identifiers.instance.format, instance_id
    )



from dynaconf.utils.boxing import DynaBox
from fastapi import Request

from mrok.controller.auth.base import UNAUTHORIZED_EXCEPTION, AuthIdentity, BaseHTTPAuthBackend
from mrok.controller.auth.registry import get_authentication_backend


class HTTPAuthManager:
    def __init__(self, auth_settings: DynaBox):
        self.auth_settings = auth_settings
        self.active_backends: list[BaseHTTPAuthBackend] = []
        self._setup_backends()

    def _setup_backends(self):
        enabled_keys = self.auth_settings.get("backends", [])

        for key in enabled_keys:
            backend_cls = get_authentication_backend(key)
            if not backend_cls:
                raise ValueError(f"Backend '{key}' is not registered.")

            specific_config = self.auth_settings.get(key, {})
            self.active_backends.append(backend_cls(specific_config))

    async def __call__(self, request: Request) -> AuthIdentity:
        for backend in self.active_backends:
            identity = await backend(request)
            if identity:
                return identity

        raise UNAUTHORIZED_EXCEPTION



import logging

import httpx
import jwt
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from fastapi.security.http import HTTPBase

from mrok.controller.auth.base import UNAUTHORIZED_EXCEPTION, AuthIdentity, BaseHTTPAuthBackend
from mrok.controller.auth.registry import register_authentication_backend

logger = logging.getLogger("mrok.controller")


@register_authentication_backend("oidc")
class OIDCJWTAuthenticationBackend(BaseHTTPAuthBackend):
    def init_scheme(self) -> HTTPBase:
        return HTTPBearer(auto_error=False)

    async def authenticate(self, credentials: HTTPAuthorizationCredentials) -> AuthIdentity | None:
        async with httpx.AsyncClient() as client:
            try:
                config_resp = await client.get(self.config.config_url)
                config_resp.raise_for_status()
                config = config_resp.json()
                issuer = config["issuer"]
                jwks_uri = config["jwks_uri"]

                jwks_resp = await client.get(jwks_uri)
                jwks_resp.raise_for_status()
                jwks = jwks_resp.json()

                header = jwt.get_unverified_header(credentials.credentials)
                kid = header["kid"]

                key_data = next((k for k in jwks["keys"] if k["kid"] == kid), None)
            except Exception:
                logger.exception("Error fetching openid-config/jwks")
                raise UNAUTHORIZED_EXCEPTION
        if key_data is None:
            logger.error("Key ID not found in JWKS")
            raise UNAUTHORIZED_EXCEPTION

        try:
            payload = jwt.decode(
                credentials.credentials,
                jwt.PyJWK(key_data),
                algorithms=[header["alg"]],
                issuer=issuer,
                audience=self.config.audience,
            )
            return AuthIdentity(
                subject=payload.get(self.config.subject_claim) or "",
                metadata=payload,
            )
        except jwt.InvalidKeyError as e:
            logger.error(f"Invalid jwt token: {e} ({credentials.credentials})")
            raise UNAUTHORIZED_EXCEPTION
        except jwt.InvalidTokenError as e:
            logger.error(f"Invalid jwt token: {e} ({credentials.credentials})")
            raise UNAUTHORIZED_EXCEPTION



import logging
from pathlib import Path
from typing import Literal

from httpcore import AsyncConnectionPool

from mrok.proxy.app import ProxyAppBase
from mrok.types.proxy import Scope

logger = logging.getLogger("mrok.agent")


TargetType = Literal["tcp", "unix"]


class SidecarProxyApp(ProxyAppBase):
    def __init__(
        self,
        target: str | Path | tuple[str, int],
        *,
        max_connections: int | None = 10,
        max_keepalive_connections: int | None = None,
        keepalive_expiry: float | None = None,
        retries: int = 0,
    ):
        self._target = target
        self._target_type, self._target_address = self._parse_target()
        super().__init__(
            max_connections=max_connections,
            max_keepalive_connections=max_keepalive_connections,
            keepalive_expiry=keepalive_expiry,
            retries=retries,
        )

    def setup_connection_pool(
        self,
        max_connections: int | None,
        max_keepalive_connections: int | None,
        keepalive_expiry: float | None,
        retries: int,
    ) -> AsyncConnectionPool:
        if self._target_type == "unix":
            return AsyncConnectionPool(
                max_connections=max_connections,
                max_keepalive_connections=max_keepalive_connections,
                keepalive_expiry=keepalive_expiry,
                retries=retries,
                uds=self._target_address,
            )
        return AsyncConnectionPool(
            max_connections=max_connections,
            max_keepalive_connections=max_keepalive_connections,
            keepalive_expiry=keepalive_expiry,
            retries=retries,
        )

    def get_upstream_base_url(self, scope: Scope) -> str:
        if self._target_type == "unix":
            return "http://localhost"
        return f"http://{self._target_address}"

    def _parse_target(self) -> tuple[TargetType, str]:
        if isinstance(self._target, Path) or (
            isinstance(self._target, str) and ":" not in self._target
        ):
            return "unix", str(self._target)

        if isinstance(self._target, str) and ":" in self._target:
            host, port = str(self._target).split(":", 1)
            host = host or "127.0.0.1"
        elif isinstance(self._target, tuple) and len(self._target) == 2:
            host = self._target[0]
            port = str(self._target[1])
        else:
            raise Exception(f"Invalid target address: {self._target}")

        return "tcp", f"{host}:{port}"



from __future__ import annotations

import asyncio
import logging
import os
import time

import psutil
from hdrh.histogram import HdrHistogram

from mrok.proxy.models import (
    DataTransferMetrics,
    ProcessMetrics,
    RequestsMetrics,
    ResponseTimeMetrics,
    WorkerMetrics,
)

logger = logging.getLogger("mrok.proxy")


def _collect_process_usage(interval: float) -> ProcessMetrics:
    proc = psutil.Process(os.getpid())
    total_cpu = 0.0
    total_mem = 0.0

    try:
        total_cpu = proc.cpu_percent(None)
    except Exception:
        total_cpu = 0.0

    if interval and interval > 0:  # pragma: no branch
        time.sleep(interval)

    try:
        total_cpu = proc.cpu_percent(None)
    except Exception:  # pragma: no cover
        total_cpu = 0.0

    try:
        total_mem = proc.memory_percent()
    except Exception:  # pragma: no cover
        total_mem = 0.0

    return ProcessMetrics(cpu=total_cpu, mem=total_mem)


async def get_process_metrics(interval: float = 0.1) -> ProcessMetrics:
    return await asyncio.to_thread(_collect_process_usage, interval)


class MetricsCollector:
    def __init__(self, worker_id: str, lowest=1, highest=60000, sigfigs=3):
        self.worker_id = worker_id
        self.total_requests = 0
        self.successful_requests = 0
        self.failed_requests = 0
        self.bytes_in = 0
        self.bytes_out = 0

        self._tick_last = time.time()
        self._tick_requests = 0

        self.hist = HdrHistogram(lowest, highest, sigfigs)

        self._lock = asyncio.Lock()

    async def on_request_start(self, scope):
        return time.perf_counter()

    async def on_request_body(self, length):
        async with self._lock:
            self.bytes_in += length

    async def on_response_start(self, status_code):
        pass  # reserved

    async def on_response_chunk(self, length):
        async with self._lock:
            self.bytes_out += length

    async def on_request_end(self, start_time, status_code):
        elapsed_ms = (time.perf_counter() - start_time) * 1000

        async with self._lock:
            self.total_requests += 1
            self._tick_requests += 1

            if status_code < 500:
                self.successful_requests += 1
            else:
                self.failed_requests += 1

            self.hist.record_value(elapsed_ms)

    async def snapshot(self) -> WorkerMetrics:
        async with self._lock:
            now = time.time()
            delta = now - self._tick_last
            rps = int(self._tick_requests / delta) if delta > 0 else 0
            data = WorkerMetrics(
                worker_id=self.worker_id,
                process=await get_process_metrics(),
                requests=RequestsMetrics(
                    rps=rps,
                    total=self.total_requests,
                    successful=self.successful_requests,
                    failed=self.failed_requests,
                ),
                data_transfer=DataTransferMetrics(
                    bytes_in=self.bytes_in,
                    bytes_out=self.bytes_out,
                ),
                response_time=ResponseTimeMetrics(
                    avg=self.hist.get_mean_value(),
                    min=self.hist.get_min_value(),
                    max=self.hist.get_max_value(),
                    p50=self.hist.get_value_at_percentile(50),
                    p90=self.hist.get_value_at_percentile(90),
                    p99=self.hist.get_value_at_percentile(99),
                ),
            )

            self._tick_last = now
            self._tick_requests = 0

            return data



import asyncio
import json
from pathlib import Path
from typing import Annotated

import typer

from mrok.cli.commands.admin.utils import parse_tags
from mrok.cli.utils import (
    validate_extension_id,
    validate_instance_id,
)
from mrok.conf import Settings, get_settings
from mrok.ziti.api import ZitiClientAPI, ZitiManagementAPI
from mrok.ziti.identities import register_identity


async def do_register(
    settings: Settings, extension_id: str, instance_id: str, tags: list[str] | None
):
    async with ZitiManagementAPI(settings) as mgmt_api, ZitiClientAPI(settings) as client_api:
        return await register_identity(
            settings, mgmt_api, client_api, extension_id, instance_id, tags=parse_tags(tags)
        )


def register(app: typer.Typer) -> None:
    settings = get_settings()

    @app.command("instance")
    def register_instance(
        ctx: typer.Context,
        extension_id: str = typer.Argument(
            ...,
            callback=validate_extension_id,
            help=f"Extension ID in the format {settings.identifiers.extension.format}",
        ),
        instance_id: str = typer.Argument(
            ...,
            callback=validate_instance_id,
            help=f"Instance ID in the format {settings.identifiers.instance.format}",
        ),
        output: Path = typer.Argument(
            ...,
            file_okay=True,
            dir_okay=False,
            writable=True,
            resolve_path=True,
            help="Output file (default: stdout)",
        ),
        tags: Annotated[
            list[str] | None,
            typer.Option(
                "--tag",
                "-t",
                help="Add tag",
                show_default=True,
            ),
        ] = None,
    ):
        """Register a new Extension Instance in OpenZiti (identity)."""
        _, identity_file = asyncio.run(do_register(ctx.obj, extension_id, instance_id, tags))
        json.dump(identity_file, output.open("w"))
        print(f"🍻 [green]Instance [bold]{instance_id}[/bold] registered.[/green]")



import asyncio

import typer

from mrok.cli.utils import validate_extension_id
from mrok.conf import Settings, get_settings
from mrok.ziti.api import ZitiManagementAPI
from mrok.ziti.services import unregister_service


async def do_unregister(settings: Settings, extension_id: str):
    async with ZitiManagementAPI(settings) as api:
        await unregister_service(settings, api, extension_id)


def register(app: typer.Typer) -> None:
    settings = get_settings()

    @app.command("extension")
    def unregister_extension(
        ctx: typer.Context,
        extension_id: str = typer.Argument(
            ...,
            callback=validate_extension_id,
            help=f"Extension ID in the format {settings.identifiers.extension.format}",
        ),
    ):
        """Unregister a new Extension in OpenZiti (service)."""
        asyncio.run(do_unregister(ctx.obj, extension_id))
        print(f"🍻 [green]Extension [bold]{extension_id}[/bold] unregistered.[/green]")



import asyncio
import select
import sys
from typing import Any

from httpcore import AsyncNetworkStream

from mrok.types.proxy import ASGIReceive


def is_readable(sock):  # pragma: no cover
    # Stolen from
    # https://github.com/python-trio/trio/blob/20ee2b1b7376db637435d80e266212a35837ddcc/trio/_socket.py#L471C1-L478C31

    # use select.select on Windows, and select.poll everywhere else
    if sys.platform == "win32":
        rready, _, _ = select.select([sock], [], [], 0)
        return bool(rready)
    p = select.poll()
    p.register(sock, select.POLLIN)
    return bool(p.poll(0))


class AIONetworkStream(AsyncNetworkStream):
    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        self._reader = reader
        self._writer = writer

    async def read(self, n: int, timeout: float | None = None) -> bytes:
        return await asyncio.wait_for(self._reader.read(n), timeout)

    async def write(self, data: bytes, timeout: float | None = None) -> None:
        self._writer.write(data)
        await asyncio.wait_for(self._writer.drain(), timeout)

    async def aclose(self) -> None:
        self._writer.close()
        await self._writer.wait_closed()

    def get_extra_info(self, info: str) -> Any:
        transport = self._writer.transport
        if info == "is_readable":
            sock = transport.get_extra_info("socket")
            return is_readable(sock)
        return transport.get_extra_info(info)


class ASGIRequestBodyStream:
    def __init__(self, receive: ASGIReceive):
        self._receive = receive
        self._more_body = True

    def __aiter__(self):
        return self

    async def __anext__(self) -> bytes:
        if not self._more_body:
            raise StopAsyncIteration

        msg = await self._receive()
        if msg["type"] == "http.request":
            chunk = msg.get("body", b"")
            self._more_body = msg.get("more_body", False)
            return chunk
        elif msg["type"] == "http.disconnect":
            raise Exception("Client disconnected.")

        raise Exception("Unexpected asgi message.")



import asyncio
import contextlib
import logging

import zmq
import zmq.asyncio

from mrok.proxy.asgi import ASGIAppWrapper
from mrok.proxy.metrics import MetricsCollector
from mrok.proxy.middleware import CaptureMiddleware, MetricsMiddleware
from mrok.proxy.models import Event, HTTPResponse, ServiceMetadata, Status
from mrok.types.proxy import ASGIApp

logger = logging.getLogger("mrok.proxy")


class EventsPublisher:
    def __init__(
        self,
        worker_id: str,
        meta: ServiceMetadata | None = None,
        events_publisher_port: int = 50000,
        events_metrics_collect_interval: float = 5.0,
    ):
        self._worker_id = worker_id
        self._meta = meta
        self._events_metrics_collect_interval = events_metrics_collect_interval
        self._publisher_port = events_publisher_port
        self._zmq_ctx = zmq.asyncio.Context()
        self._publisher = self._zmq_ctx.socket(zmq.PUB)
        self._metrics_collector = MetricsCollector(self._worker_id)
        self._publish_task = None

    async def on_startup(self):
        self._publisher.connect(f"tcp://localhost:{self._publisher_port}")
        self._publish_task = asyncio.create_task(self.publish_metrics_event())
        logger.info(f"Events publishing for worker {self._worker_id} started")

    async def on_shutdown(self):
        self._publish_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await self._publish_task
        self._publisher.close()
        self._zmq_ctx.term()
        logger.info(f"Events publishing for worker {self._worker_id} stopped")

    async def publish_metrics_event(self):
        while True:
            snap = await self._metrics_collector.snapshot()
            event = Event(type="status", data=Status(meta=self._meta, metrics=snap))
            await self._publisher.send_string(event.model_dump_json())
            await asyncio.sleep(self._events_metrics_collect_interval)

    async def publish_response_event(self, response: HTTPResponse):
        event = Event(type="response", data=response)
        await self._publisher.send_string(event.model_dump_json())  # type: ignore[attr-defined]

    def setup_middleware(self, app: ASGIAppWrapper):
        app.add_middleware(CaptureMiddleware, self.publish_response_event)
        app.add_middleware(MetricsMiddleware, self._metrics_collector)  # type: ignore

    @contextlib.asynccontextmanager
    async def lifespan(self, app: ASGIApp):
        await self.on_startup()  # type: ignore
        yield
        await self.on_shutdown()  # type: ignore



from pathlib import Path
from typing import Annotated

import typer

from mrok import frontend
from mrok.cli.utils import number_of_workers

default_workers = number_of_workers()


def register(app: typer.Typer) -> None:
    @app.command("run")
    def run_frontend(
        ctx: typer.Context,
        identity_file: Path = typer.Argument(
            ...,
            help="Identity json file",
        ),
        host: Annotated[
            str,
            typer.Option(
                "--host",
                "-h",
                help="Host to bind to. Default: 127.0.0.1",
                show_default=True,
            ),
        ] = "127.0.0.1",
        port: Annotated[
            int,
            typer.Option(
                "--port",
                "-p",
                help="Port to bind to. Default: 8000",
                show_default=True,
            ),
        ] = 8000,
        workers: Annotated[
            int,
            typer.Option(
                "--workers",
                "-w",
                help=f"Number of workers. Default: {default_workers}",
                show_default=True,
            ),
        ] = default_workers,
        max_connections: Annotated[
            int,
            typer.Option(
                "--max-pool-connections",
                help=(
                    "The maximum number of concurrent HTTP connections that "
                    "the pool should allow. Any attempt to send a request on a pool that "
                    "would exceed this amount will block until a connection is available."
                ),
                show_default=True,
            ),
        ] = 1000,
        max_keepalive_connections: Annotated[
            int | None,
            typer.Option(
                "--max-pool-keepalive-connections",
                help=(
                    "The maximum number of idle HTTP connections "
                    "that will be maintained in the pool."
                ),
                show_default=True,
            ),
        ] = 100,
        keepalive_expiry: Annotated[
            float | None,
            typer.Option(
                "--max-pool-keepalive-expiry",
                help=(
                    "The duration in seconds that an idle HTTP connection "
                    "may be maintained for before being expired from the pool."
                ),
                show_default=True,
            ),
        ] = 300,
    ):
        """Run the mrok frontend with Gunicorn and Uvicorn workers."""
        frontend.run(
            identity_file,
            host,
            port,
            workers,
            max_connections=max_connections,
            max_keepalive_connections=max_keepalive_connections,
            keepalive_expiry=keepalive_expiry,
        )



from pytest_mock import MockerFixture

from mrok.agent.sidecar import main


def test_sidecar_agent(mocker: MockerFixture):
    mocked_app = mocker.MagicMock()
    mocked_app_ctor = mocker.patch(
        "mrok.agent.sidecar.main.SidecarProxyApp",
        return_value=mocked_app,
    )
    agent = main.SidecarAgent(
        "identity.json",
        ":8000",
        server_workers=2,
        events_publishers_port=2000,
        events_subscribers_port=3000,
    )
    assert agent.server_reload is False
    assert agent.get_asgi_app() == mocked_app
    mocked_app_ctor.assert_called_once_with(
        ":8000",
        max_connections=10,
        max_keepalive_connections=None,
        keepalive_expiry=None,
        retries=0,
    )


def test_run(mocker: MockerFixture):
    mocked_agent = mocker.MagicMock()
    mocked_agent_ctor = mocker.patch(
        "mrok.agent.sidecar.main.SidecarAgent",
        return_value=mocked_agent,
    )

    main.run(
        "ziti-identity.json",
        "target-addr",
        server_workers=10,
        upstream_max_connections=15,
        upstream_max_keepalive_connections=3,
        upstream_keepalive_expiry=100,
        upstream_max_connect_retries=6,
        events_publishers_port=4000,
        events_subscribers_port=5000,
    )

    mocked_agent_ctor.assert_called_once_with(
        "ziti-identity.json",
        "target-addr",
        events_enabled=True,
        server_workers=10,
        upstream_max_connections=15,
        upstream_max_keepalive_connections=3,
        upstream_max_connect_retries=6,
        upstream_keepalive_expiry=100,
        events_publishers_port=4000,
        events_subscribers_port=5000,
        events_metrics_collect_interval=5.0,
        server_backlog=2048,
        server_limit_concurrency=None,
        server_limit_max_requests=None,
        server_timeout_keep_alive=5,
        ziti_load_timeout_ms=5000,
    )
    mocked_agent.run.assert_called_once()



import logging
import os
import signal
import threading
import time
from abc import ABC, abstractmethod
from pathlib import Path
from uuid import uuid4

import zmq
from watchfiles import watch
from watchfiles.filters import PythonFilter
from watchfiles.run import CombinedProcess, start_process

from mrok.conf import get_settings
from mrok.logging import setup_logging
from mrok.proxy.worker import Worker
from mrok.types.proxy import ASGIApp

logger = logging.getLogger("mrok.agent")

MONITOR_THREAD_JOIN_TIMEOUT = 5
MONITOR_THREAD_CHECK_DELAY = 1


def print_path(path):
    try:
        return f'"{path.relative_to(Path.cwd())}"'
    except ValueError:
        return f'"{path}"'


def start_uvicorn_worker(
    worker_id: str,
    app: ASGIApp | str,
    identity_file: str,
    *,
    ziti_load_timeout_ms: int = 5000,
    server_backlog: int = 2048,
    server_timeout_keep_alive: int = 5,
    server_limit_concurrency: int | None = None,
    server_limit_max_requests: int | None = None,
    events_enabled: bool = True,
    events_pub_port: int = 5000,
    events_metrics_collect_interval: float = 5.0,
):
    import sys

    sys.path.insert(0, os.getcwd())

    worker = Worker(
        worker_id,
        app,
        identity_file,
        ziti_load_timeout_ms=ziti_load_timeout_ms,
        server_backlog=server_backlog,
        server_timeout_keep_alive=server_timeout_keep_alive,
        server_limit_concurrency=server_limit_concurrency,
        server_limit_max_requests=server_limit_max_requests,
        events_enabled=events_enabled,
        events_publisher_port=events_pub_port,
        events_metrics_collect_interval=events_metrics_collect_interval,
    )
    worker.run()


MONITOR_THREAD_ERROR_DELAY = 3


def start_events_router(events_pub_port: int, events_sub_port: int):
    setup_logging(get_settings())
    context = zmq.Context()
    frontend = context.socket(zmq.XSUB)
    frontend.bind(f"tcp://localhost:{events_pub_port}")
    backend = context.socket(zmq.XPUB)
    backend.bind(f"tcp://localhost:{events_sub_port}")

    try:
        logger.info(f"Events router process started: {os.getpid()}")
        zmq.proxy(frontend, backend)
    except KeyboardInterrupt:  # pragma: no cover
        pass
    finally:
        frontend.close()
        backend.close()
        context.term()


class MasterBase(ABC):
    def __init__(
        self,
        identity_file: str,
        *,
        ziti_load_timeout_ms: int = 5000,
        server_workers: int = 4,
        server_reload: bool = False,
        server_backlog: int = 2048,
        server_timeout_keep_alive: int = 5,
        server_limit_concurrency: int | None = None,
        server_limit_max_requests: int | None = None,
        events_enabled: bool = True,
        events_pub_port: int = 50000,
        events_sub_port: int = 50001,
        events_metrics_collect_interval: float = 5.0,
    ):
        self.identity_file = identity_file
        self.workers = server_workers
        self.server_reload = server_reload
        self.events_enabled = events_enabled
        self.events_pub_port = events_pub_port
        self.events_sub_port = events_sub_port
        self.events_metrics_collect_interval = events_metrics_collect_interval
        self.worker_identifiers = [str(uuid4()) for _ in range(server_workers)]
        self.worker_processes: dict[str, CombinedProcess] = {}
        self.zmq_pubsub_router_process = None
        self.monitor_thread = threading.Thread(target=self.monitor_workers, daemon=True)
        self.stop_event = threading.Event()
        self.pause_event = threading.Event()
        self.watch_filter = PythonFilter(ignore_paths=None)
        self.watcher = watch(
            Path.cwd(),
            watch_filter=self.watch_filter,
            stop_event=self.stop_event,
            yield_on_timeout=True,
        )
        self.ziti_load_timeout_ms = ziti_load_timeout_ms
        self.server_backlog = server_backlog
        self.timeout_keep_alive = server_timeout_keep_alive
        self.limit_concurrency = server_limit_concurrency
        self.limit_max_requests = server_limit_max_requests
        self.setup_signals_handler()

    @abstractmethod
    def get_asgi_app(self):
        raise NotImplementedError()

    def setup_signals_handler(self):
        for sig in (signal.SIGINT, signal.SIGTERM):
            signal.signal(sig, self.handle_signal)

    def handle_signal(self, *args, **kwargs):
        self.stop_event.set()

    def start_worker(self, worker_id: str):
        """Start a single worker process"""

        p = start_process(
            start_uvicorn_worker,
            "function",
            (
                worker_id,
                self.get_asgi_app(),
                self.identity_file,
            ),
            {
                "ziti_load_timeout_ms": self.ziti_load_timeout_ms,
                "server_backlog": self.server_backlog,
                "server_timeout_keep_alive": self.timeout_keep_alive,
                "server_limit_concurrency": self.limit_concurrency,
                "server_limit_max_requests": self.limit_max_requests,
                "events_enabled": self.events_enabled,
                "events_pub_port": self.events_pub_port,
                "events_metrics_collect_interval": self.events_metrics_collect_interval,
            },
        )
        logger.info(f"Worker {worker_id} [{p.pid}] started")
        return p

    def start_events_router(self):
        self.zmq_pubsub_router_process = start_process(
            start_events_router,
            "function",
            (
                self.events_pub_port,
                self.events_sub_port,
            ),
            None,
        )

    def start_workers(self):
        for i in range(self.workers):
            worker_id = self.worker_identifiers[i]
            p = self.start_worker(worker_id)
            self.worker_processes[worker_id] = p

    def start(self):
        self.start_events_router()
        self.start_workers()
        self.monitor_thread.start()
        self.pause_event.set()

    def stop_events_router(self):
        self.zmq_pubsub_router_process.stop(sigint_timeout=5, sigkill_timeout=1)

    def stop_workers(self):
        for process in self.worker_processes.values():
            if process.is_alive():  # pragma: no branch
                process.stop(sigint_timeout=5, sigkill_timeout=1)
        self.worker_processes.clear()

    def stop(self):
        self.pause_event.clear()
        if self.monitor_thread.is_alive():  # pragma: no branch
            self.monitor_thread.join(timeout=MONITOR_THREAD_JOIN_TIMEOUT)
        self.stop_workers()
        self.stop_events_router()

    def restart(self):
        self.pause_event.clear()
        self.stop_workers()
        self.start_workers()
        self.pause_event.set()

    def monitor_workers(self):
        while not self.stop_event.is_set():
            try:
                self.pause_event.wait()
                for worker_id, process in self.worker_processes.items():
                    if not process.is_alive():
                        logger.warning(f"Worker {worker_id} [{process.pid}] died unexpectedly")
                        process.stop(sigint_timeout=1, sigkill_timeout=1)
                        new_process = self.start_worker(worker_id)
                        self.worker_processes[worker_id] = new_process
                        logger.info(
                            f"Restarted worker {worker_id} [{process.pid}] -> [{new_process.pid}]"
                        )

                time.sleep(MONITOR_THREAD_CHECK_DELAY)

            except Exception as e:
                logger.error(f"Error in worker monitoring: {e}")
                time.sleep(MONITOR_THREAD_ERROR_DELAY)

    def __iter__(self):
        return self

    def __next__(self):
        changes = next(self.watcher)
        if changes:
            return list({Path(change[1]) for change in changes})
        return None

    def run(self):
        setup_logging(get_settings())
        logger.info(f"Master process started: {os.getpid()}")
        self.start()
        try:
            if self.server_reload:
                for files_changed in self:
                    if files_changed:
                        logger.warning(
                            f"{', '.join(map(print_path, files_changed))} changed, reloading...",
                        )
                        self.restart()
            else:
                self.stop_event.wait()
        finally:
            self.stop()



from collections.abc import Iterator
from contextlib import AsyncExitStack, asynccontextmanager
from typing import Any, ParamSpec, Protocol

from mrok.types.proxy import ASGIApp, ASGIReceive, ASGISend, Lifespan, Scope

P = ParamSpec("P")


class ASGIMiddleware(Protocol[P]):
    def __call__(
        self, app: ASGIApp, /, *args: P.args, **kwargs: P.kwargs
    ) -> ASGIApp: ...  # pragma: no cover


class Middleware:
    def __init__(self, cls: ASGIMiddleware[P], *args: P.args, **kwargs: P.kwargs) -> None:
        self.cls = cls
        self.args = args
        self.kwargs = kwargs

    def __iter__(self) -> Iterator[Any]:
        as_tuple = (self.cls, self.args, self.kwargs)
        return iter(as_tuple)


class ASGIAppWrapper:
    def __init__(
        self,
        app: ASGIApp,
        lifespan: Lifespan | None = None,
    ) -> None:
        self.app = app
        self.lifespan = lifespan
        self.middlware: list[Middleware] = []
        self.middleare_stack: ASGIApp | None = None

    def add_middleware(self, cls: ASGIMiddleware[P], *args: P.args, **kwargs: P.kwargs):
        self.middlware.insert(0, Middleware(cls, *args, **kwargs))

    def build_middleware_stack(self):
        app = self.app
        for cls, args, kwargs in reversed(self.middlware):
            app = cls(app, *args, **kwargs)
        return app

    def get_starlette_lifespan(self):
        router = getattr(self.app, "router", None)
        if router is None:
            return None
        return getattr(router, "lifespan_context", None)

    @asynccontextmanager
    async def merge_lifespan(self, app: ASGIApp):
        async with AsyncExitStack() as stack:
            state: dict[Any, Any] = {}
            if self.lifespan is not None:
                outer_state = await stack.enter_async_context(self.lifespan(app))
                state.update(outer_state or {})
            starlette_lifespan = self.get_starlette_lifespan()
            if starlette_lifespan is not None:
                inner_state = await stack.enter_async_context(starlette_lifespan(app))
                state.update(inner_state or {})
            yield state

    async def handle_lifespan(self, scope: Scope, receive: ASGIReceive, send: ASGISend) -> None:
        started = False
        app: Any = scope.get("app")
        await receive()
        try:
            async with self.merge_lifespan(app) as state:
                if state:
                    if "state" not in scope:
                        raise RuntimeError('"state" is unsupported by the current ASGI Server.')
                    scope["state"].update(state)
                await send({"type": "lifespan.startup.complete"})
                started = True
                await receive()
        except Exception as e:  # pragma: no cover
            if started:
                await send({"type": "lifespan.shutdown.failed", "message": str(e)})
            else:
                await send({"type": "lifespan.startup.failed", "message": str(e)})
            raise
        else:
            await send({"type": "lifespan.shutdown.complete"})

    async def __call__(self, scope: Scope, receive: ASGIReceive, send: ASGISend) -> None:
        if self.middleare_stack is None:  # pragma: no branch
            self.middleware_stack = self.build_middleware_stack()
        if scope["type"] == "lifespan":
            scope["app"] = self
            await self.handle_lifespan(scope, receive, send)
            return

        await self.middleware_stack(scope, receive, send)



from __future__ import annotations

from collections.abc import Awaitable, Callable, Coroutine, Mapping, MutableMapping
from contextlib import AbstractAsyncContextManager
from typing import Any, Never

from mrok.proxy.models import HTTPResponse

Scope = MutableMapping[str, Any]
Message = MutableMapping[str, Any]

ASGIReceive = Callable[[], Awaitable[Message]]
ASGISend = Callable[[Message], Awaitable[None]]
ASGIApp = Callable[[Scope, ASGIReceive, ASGISend], Awaitable[None]]
StatelessLifespan = Callable[[ASGIApp], AbstractAsyncContextManager[None]]
StatefulLifespan = Callable[[ASGIApp], AbstractAsyncContextManager[Mapping[str, Any]]]
Lifespan = StatelessLifespan | StatefulLifespan

LifespanCallback = Callable[[], Awaitable[None]]
ResponseCompleteCallback = Callable[[HTTPResponse], Coroutine[Any, Any, Never]]



from pytest_mock import MockerFixture

from mrok.proxy.asgi import ASGIAppWrapper
from mrok.proxy.middleware import CaptureMiddleware, MetricsMiddleware
from mrok.proxy.worker import Worker
from tests.types import SettingsFactory


def test_setup_app(
    mocker: MockerFixture,
    ziti_identity_file: str,
):
    m_app = mocker.AsyncMock()
    worker = Worker(
        "my-worker-id",
        m_app,
        ziti_identity_file,
    )
    app = worker.setup_app()
    assert isinstance(app, ASGIAppWrapper)
    assert worker._event_publisher is not None
    assert app.lifespan == worker._event_publisher.lifespan
    assert app.middlware[0].cls == MetricsMiddleware
    assert app.middlware[0].args[0] == worker._event_publisher._metrics_collector
    assert app.middlware[1].cls == CaptureMiddleware
    assert app.middlware[1].args[0] == worker._event_publisher.publish_response_event


def test_setup_app_events_disabled(
    mocker: MockerFixture,
    ziti_identity_file: str,
):
    m_app = mocker.AsyncMock()
    worker = Worker(
        "my-worker-id",
        m_app,
        ziti_identity_file,
        server_backlog=2048,
        server_limit_concurrency=None,
        server_limit_max_requests=None,
        server_timeout_keep_alive=5,
        ziti_load_timeout_ms=5000,
        events_enabled=False,
    )
    app = worker.setup_app()
    assert isinstance(app, ASGIAppWrapper)
    assert app.lifespan is None

    assert app.middlware == []


def test_run(
    mocker: MockerFixture,
    ziti_identity_file: str,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocker.patch("mrok.proxy.worker.get_settings", return_value=settings)
    m_setup_logging = mocker.patch("mrok.proxy.worker.setup_logging")

    m_mrokconfig = mocker.MagicMock()
    m_mrokconfig_ctor = mocker.patch("mrok.proxy.worker.BackendConfig", return_value=m_mrokconfig)

    m_server = mocker.MagicMock()
    m_server_ctor = mocker.patch("mrok.proxy.worker.Server", return_value=m_server)
    m_app = mocker.MagicMock()
    mocker.patch.object(Worker, "setup_app", return_value=m_app)

    worker = Worker(
        "my-worker-id",
        m_app,
        ziti_identity_file,
        ziti_load_timeout_ms=5000,
        server_backlog=2048,
        server_timeout_keep_alive=5,
        server_limit_concurrency=None,
        server_limit_max_requests=None,
    )
    worker.run()

    m_setup_logging.assert_called_once_with(settings)
    m_mrokconfig_ctor.assert_called_once_with(
        m_app,
        ziti_identity_file,
        ziti_load_timeout_ms=5000,
        backlog=2048,
        timeout_keep_alive=5,
        limit_concurrency=None,
        limit_max_requests=None,
    )
    m_server_ctor.assert_called_once_with(m_mrokconfig)
    m_server.run.assert_called_once()



import asyncio

import pytest
from pytest_mock import MockerFixture
from textual.pilot import Pilot

from mrok.agent.devtools.inspector.app import MIN_COLS, MIN_ROWS, InspectorApp, module_main
from tests.types import ResponseEventFactory, SnapCompare, StatusEventFactory, ZMQPublisher


def test_app_get_headers(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, response_event_factory())
        await asyncio.to_thread(s.send_json, status_event_factory())
        await pilot.click(offset=(5, 9))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_get_payload(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, response_event_factory())
        await asyncio.to_thread(s.send_json, status_event_factory())
        await pilot.click(offset=(5, 9))
        await pilot.click(offset=(13, 20))
        await pilot.click(offset=(2, 22))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_get_preview(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, response_event_factory())
        await asyncio.to_thread(s.send_json, status_event_factory())
        await pilot.click(offset=(5, 9))
        await pilot.click(offset=(22, 20))
        await pilot.click(offset=(2, 22))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_get_raw_response(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher

    status = status_event_factory(process_cpu=32.2, process_mem=100)

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, status)
        await asyncio.to_thread(s.send_json, response_event_factory())
        await pilot.click(offset=(5, 9))
        await pilot.click(offset=(31, 20))
        await pilot.click(offset=(2, 22))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_get_preview_binary(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher
    response_event = response_event_factory(
        response_headers={"content-type": "application/octet-stream"},
        response_body=b"blablabla",
    )

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, response_event)
        await asyncio.to_thread(s.send_json, status_event_factory())
        await pilot.click(offset=(5, 9))
        await pilot.click(offset=(22, 20))
        await pilot.click(offset=(2, 22))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_get_raw_response_binary(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher

    response_event = response_event_factory(
        response_headers={"content-type": "application/octet-stream"},
        response_body=b"blablabla",
    )

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, status_event_factory())
        await asyncio.to_thread(s.send_json, response_event)
        await pilot.click(offset=(5, 9))
        await pilot.click(offset=(31, 20))
        await pilot.click(offset=(2, 22))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_post_headers(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher

    response_event = response_event_factory(
        method="POST",
        request_headers={
            "accept": "application/json",
            "content-type": "application/json",
        },
        request_body=b'{"request": "body"}',
    )

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, status_event_factory())
        await asyncio.to_thread(s.send_json, response_event)
        await pilot.click(offset=(5, 9))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_post_json_body(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher

    response_event = response_event_factory(
        method="POST",
        request_headers={
            "accept": "application/json",
            "content-type": "application/json",
        },
        request_body=b'{"request": "body"}',
    )

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, status_event_factory())
        await asyncio.to_thread(s.send_json, response_event)
        await pilot.click(offset=(5, 9))
        await pilot.click(offset=(13, 20))
        await pilot.click(offset=(2, 22))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_post_preview_multipart(
    multipart_body: tuple[str, bytes],
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher
    boundary, body = multipart_body

    response_event = response_event_factory(
        method="POST",
        request_headers={
            "accept": "application/json",
            "content-type": f"multipart/form-data; boundary={boundary}",
        },
        request_body=body,
    )

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, status_event_factory())
        await asyncio.to_thread(s.send_json, response_event)
        await pilot.click(offset=(5, 9))
        await pilot.click(offset=(13, 20))
        await pilot.click(offset=(2, 22))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_post_preview_body_truncated(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher

    response_event = response_event_factory(
        method="POST",
        request_headers={
            "accept": "application/json",
            "content-type": "application/octet-stream",
        },
        request_body=b"These are a lot of random bytes",
        request_truncated=True,
    )

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, status_event_factory())
        await asyncio.to_thread(s.send_json, response_event)
        await pilot.click(offset=(5, 9))
        await pilot.click(offset=(13, 20))
        await pilot.click(offset=(2, 22))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_delete_headers(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
):
    s, port = zmq_publisher

    response_event = response_event_factory(
        method="DELETE",
    )

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, status_event_factory())
        await asyncio.to_thread(s.send_json, response_event)
        await pilot.click(offset=(5, 9))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


@pytest.mark.parametrize(
    "response_status",
    [400, 500],
)
def test_app_non_200_headers(
    response_event_factory: ResponseEventFactory,
    status_event_factory: StatusEventFactory,
    zmq_publisher: ZMQPublisher,
    snap_compare: SnapCompare,
    response_status: int,
):
    s, port = zmq_publisher

    response_event = response_event_factory(
        method="POST", request_querystring=b"foo=bar&goo=1", response_status=response_status
    )

    async def run_before(pilot: Pilot):
        await asyncio.to_thread(s.send_json, status_event_factory())
        await asyncio.to_thread(s.send_json, response_event)
        await pilot.click(offset=(5, 9))

    assert snap_compare(
        InspectorApp(port), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


def test_app_hide_metrics(
    snap_compare: SnapCompare,
):
    async def run_before(pilot: Pilot):
        await pilot.press("m")

    assert snap_compare(
        InspectorApp(2333), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS)
    )


@pytest.mark.parametrize(
    ("w", "h"),
    [
        (1, 0),
        (0, 1),
        (1, 1),
    ],
)
def test_app_terminal_too_small(
    snap_compare: SnapCompare,
    w: int,
    h: int,
):
    assert snap_compare(InspectorApp(9332), terminal_size=(MIN_COLS - w, MIN_ROWS - h))


def test_app_terminal_too_small_dismiss(
    snap_compare: SnapCompare,
):
    async def run_before(pilot: Pilot):
        await pilot.click("#dismiss")

    assert snap_compare(
        InspectorApp(9332), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS - 15)
    )


def test_app_terminal_too_small_resize_auto_dismiss(
    snap_compare: SnapCompare,
):
    async def run_before(pilot: Pilot):
        await pilot.resize_terminal(MIN_COLS, MIN_ROWS)

    assert snap_compare(
        InspectorApp(9332), run_before=run_before, terminal_size=(MIN_COLS, MIN_ROWS - 15)
    )


@pytest.mark.parametrize(
    ("argv", "expected_port"),
    [
        ([], 50001),
        (["-p", "4949"], 4949),
        (["--subscriber-port", "7777"], 7777),
    ],
)
def test_module_main(
    mocker: MockerFixture,
    argv: list,
    expected_port: int,
):
    m_app = mocker.MagicMock()
    m_ctor = mocker.patch("mrok.agent.devtools.inspector.app.InspectorApp", return_value=m_app)
    module_main(argv)
    m_ctor.assert_called_once_with(expected_port)
    m_app.run.assert_called_once()



import pytest

from mrok.agent.devtools.inspector.utils import (
    CONTENT_TYPE_TO_LANGUAGE,
    build_tree,
    get_highlighter_language_by_content_type,
    hexdump,
    humanize_bytes,
    parse_content_type,
    parse_form_data,
)


@pytest.mark.parametrize(
    ("content_type_string", "expected"),
    [
        ("application/json", ("application/json", "utf-8", None)),
        ("application/json; charset=latin1", ("application/json", "latin1", None)),
        (
            "multipart/form-data; boundary=---my-boundary",
            ("multipart/form-data", None, "---my-boundary"),
        ),
        ("application/json; invalid", ("application/json", "utf-8", None)),
        ("application/octet-stream; invalid", ("application/octet-stream", None, None)),
        ("application/octet-stream; other=whatever", ("application/octet-stream", None, None)),
        ("text/plain; other=whatever", ("text/plain", "utf-8", None)),
    ],
)
def test_parse_content_type(content_type_string: str, expected: tuple[str, str | None, str | None]):
    content_type_info = parse_content_type(content_type_string)
    assert content_type_info.content_type == expected[0]
    assert content_type_info.charset == expected[1]
    assert content_type_info.boundary == expected[2]


def test_parse_form_data(multipart_body: tuple[str, bytes]):
    boundary, body = multipart_body

    result = list(parse_form_data(body, boundary))

    assert len(result) == 4
    assert result[0] == ("username", "john_doe")
    assert result[1] == ("email", "john@example.com")
    assert result[2] == ("message", "Hello, this is a test message!")
    assert result[3] == ("file", "<binary>")


def test_build_tree():
    class FakeNode:
        def __init__(self, data):
            self.data = data
            self.children = []

        def add(self, data):
            child = FakeNode(data)
            self.children.append(child)
            return child

    def dump_node(node):
        return (
            node.data,
            [dump_node(child) for child in node.children],
        )

    data = {
        "a": 1,
        "b": [2, {"c": 3}],
    }
    root = FakeNode("root")
    build_tree(root, data)

    assert dump_node(root) == (
        "root",
        [
            ("a", [("1", [])]),
            (
                "b",
                [
                    ("[0]", [("2", [])]),
                    ("[1]", [("c", [("3", [])])]),
                ],
            ),
        ],
    )


def test_hexdump():
    data = b"Hello World"
    result = hexdump(data, width=16)
    assert "48 65 6c 6c 6f 20 57 6f 72 6c 64" in result
    assert "Hello World" in result


def test_hexdump_non_printable():
    data = b"\x00\x01\x02ABC"
    result = hexdump(data)
    assert "00 01 02 41 42 43" in result
    assert "...ABC" in result


def test_hexdump_custom_width():
    data = b"123456"
    result = hexdump(data, width=3)
    lines = result.split("\n")
    assert len(lines) == 2
    assert lines[0].strip().endswith("123")
    assert lines[1].strip().endswith("456")


def test_hexdump_empty_input():
    assert hexdump(b"") == ""


@pytest.mark.parametrize(
    ("input_bytes", "expected_hex"),
    [
        (b"\xff", "ff"),
        (b"\x00", "00"),
        (b"\x20", "20"),
    ],
)
def test_hexdump_single_bytes(input_bytes, expected_hex):
    assert expected_hex in hexdump(input_bytes)


@pytest.mark.parametrize(
    ("input_bytes", "expected_value", "expected_unit"),
    [
        (0, 0.0, "B"),
        (1, 1.0, "B"),
        (999, 999.0, "B"),
        (1000, 1000.0, "B"),
        (1023, 1023.0, "B"),
        (1024, 1.0, "KiB"),
        (1025, 1.0, "KiB"),
        (1536, 1.5, "KiB"),
        (2047, 2.0, "KiB"),
        (5242880, 5.0, "MiB"),
        (1024**3 - 1, 1023.99, "MiB"),
        (1024**3, 1.0, "GiB"),
        (1024**4, 1.0, "TiB"),
        (1024**5, 1.0, "PiB"),
        (1024**6, 1024.0, "PiB"),
        (99_999_999_999_999, 90.95, "TiB"),
        (1_099_511_627_776, 1.0, "TiB"),
    ],
)
def test_humanize_bytes(input_bytes, expected_value, expected_unit):
    value, unit = humanize_bytes(input_bytes)
    assert value == pytest.approx(expected_value, abs=0.01)
    assert unit == expected_unit


def test_humanize_bytes_negative_value():
    with pytest.raises(ValueError, match="num_bytes must be non-negative"):
        humanize_bytes(-1)


@pytest.mark.parametrize(
    ("content_type", "language"), list(CONTENT_TYPE_TO_LANGUAGE.items()) + [("invalid", None)]
)
def test_get_highlighter_language_by_content_type(content_type: str, language: str):
    assert get_highlighter_language_by_content_type(content_type) == language



import asyncio

import pytest
from pytest_mock import MockerFixture

from mrok.proxy.stream import AIONetworkStream, ASGIRequestBodyStream


@pytest.mark.asyncio
async def test_aio_network_stream_read(mocker: MockerFixture):
    m_reader = mocker.AsyncMock()
    m_reader.read.return_value = b"OK"

    aions = AIONetworkStream(m_reader, mocker.MagicMock())
    read_bytes = await aions.read(2)
    assert read_bytes == b"OK"
    m_reader.read.assert_awaited_once_with(2)


@pytest.mark.asyncio
async def test_aio_network_stream_read_timeout(mocker: MockerFixture):
    class FakeReader:
        async def read(self, n: int = -1):
            await asyncio.sleep(0.2)

    aions = AIONetworkStream(FakeReader(), mocker.MagicMock())  # type: ignore[arg-type]
    with pytest.raises(asyncio.TimeoutError):
        await aions.read(5, timeout=0.1)


@pytest.mark.asyncio
async def test_aio_network_stream_write(mocker: MockerFixture):
    m_writer = mocker.AsyncMock()
    m_writer.write = mocker.MagicMock()

    aions = AIONetworkStream(mocker.MagicMock(), m_writer)
    await aions.write(b"OK")
    m_writer.write.assert_called_once_with(b"OK")
    m_writer.drain.assert_awaited_once()


@pytest.mark.asyncio
async def test_aio_network_stream_write_timeout(mocker: MockerFixture):
    class FakeWriter:
        def write(self, data: bytes):
            pass

        async def drain(self):
            await asyncio.sleep(0.2)

    aions = AIONetworkStream(mocker.MagicMock(), FakeWriter())  # type: ignore[arg-type]
    with pytest.raises(asyncio.TimeoutError):
        await aions.write(b"OK", timeout=0.1)


@pytest.mark.asyncio
async def test_aio_network_stream_close(mocker: MockerFixture):
    m_writer = mocker.AsyncMock()
    m_writer.close = mocker.MagicMock()

    aions = AIONetworkStream(mocker.MagicMock(), m_writer)

    await aions.aclose()
    m_writer.close.assert_called_once()
    m_writer.wait_closed.assert_awaited_once()


@pytest.mark.parametrize("readable", [True, False])
def test_aio_network_stream_extra_info_is_readable(
    mocker: MockerFixture,
    readable: bool,
):
    m_sock = mocker.MagicMock()
    m_writer = mocker.MagicMock()
    m_writer.transport = mocker.MagicMock()
    m_writer.transport.get_extra_info.return_value = m_sock

    m_is_readable = mocker.patch("mrok.proxy.stream.is_readable", return_value=readable)

    aions = AIONetworkStream(mocker.MagicMock(), m_writer)

    assert aions.get_extra_info("is_readable") is readable
    m_is_readable.assert_called_once_with(m_sock)
    m_writer.transport.get_extra_info.assert_called_once_with("socket")


def test_aio_network_stream_extra_info_other(
    mocker: MockerFixture,
):
    m_writer = mocker.MagicMock()
    m_writer.transport = mocker.MagicMock()
    m_writer.transport.get_extra_info.return_value = "whatever"

    aions = AIONetworkStream(mocker.MagicMock(), m_writer)

    assert aions.get_extra_info("other_extra") == "whatever"
    m_writer.transport.get_extra_info.assert_called_once_with("other_extra")


@pytest.mark.asyncio
async def test_asgi_request_body_stream():
    messages = [
        {
            "type": "http.request",
            "body": b"first chunk",
            "more_body": True,
        },
        {
            "type": "http.request",
            "body": b"second chunk",
            "more_body": False,
        },
    ]

    msg_iter = iter(messages)

    async def receive():
        await asyncio.sleep(0)
        return next(msg_iter)

    stream = ASGIRequestBodyStream(receive)

    chunks = [chunk async for chunk in stream]
    assert chunks == [b"first chunk", b"second chunk"]


@pytest.mark.asyncio
async def test_asgi_request_body_stream_disconnect():
    messages = [
        {
            "type": "http.disconnect",
        },
    ]

    msg_iter = iter(messages)

    async def receive():
        await asyncio.sleep(0)
        return next(msg_iter)

    stream = ASGIRequestBodyStream(receive)

    with pytest.raises(Exception) as cv:
        await anext(stream)

    assert str(cv.value) == "Client disconnected."


@pytest.mark.asyncio
async def test_asgi_request_body_unknown_msg():
    messages = [
        {
            "type": "whatever",
        },
    ]

    msg_iter = iter(messages)

    async def receive():
        await asyncio.sleep(0)
        return next(msg_iter)

    stream = ASGIRequestBodyStream(receive)

    with pytest.raises(Exception) as cv:
        await anext(stream)

    assert str(cv.value) == "Unexpected asgi message."



import pytest
from pytest_mock import MockerFixture

from mrok.agent.sidecar.app import SidecarProxyApp


@pytest.mark.parametrize(
    ("target", "addr"),
    [
        ("127.0.0.1:1234", "127.0.0.1:1234"),
        (":8282", "127.0.0.1:8282"),
        (("localhost", 1838), "localhost:1838"),
    ],
)
def test_init(
    mocker: MockerFixture,
    target: str | tuple[str, int],
    addr: str,
):
    m_async_pool_ctor = mocker.patch("mrok.agent.sidecar.app.AsyncConnectionPool")
    app = SidecarProxyApp(
        target,
        max_connections=5000,
        max_keepalive_connections=5,
        keepalive_expiry=60,
        retries=1,
    )

    m_async_pool_ctor.assert_called_once_with(
        max_connections=5000,
        max_keepalive_connections=5,
        keepalive_expiry=60,
        retries=1,
    )

    assert app._target_type == "tcp"
    assert app._target_address == addr


def test_init_uds(
    mocker: MockerFixture,
):
    m_async_pool_ctor = mocker.patch("mrok.agent.sidecar.app.AsyncConnectionPool")
    app = SidecarProxyApp(
        "/path/to/proxy.sock",
        max_connections=5000,
        max_keepalive_connections=5,
        keepalive_expiry=60,
        retries=1,
    )

    m_async_pool_ctor.assert_called_once_with(
        max_connections=5000,
        max_keepalive_connections=5,
        keepalive_expiry=60,
        retries=1,
        uds="/path/to/proxy.sock",
    )

    assert app._target_type == "unix"
    assert app._target_address == "/path/to/proxy.sock"


@pytest.mark.parametrize(
    "target",
    [
        ("a",),
        ("a", "b", "c"),
    ],
)
def test_init_invalid_target(
    mocker: MockerFixture,
    target: tuple,
):
    with pytest.raises(Exception) as cv:
        SidecarProxyApp(
            target,
            max_connections=5000,
            max_keepalive_connections=5,
            keepalive_expiry=60,
            retries=1,
        )
    assert str(cv.value) == f"Invalid target address: {target}"


def test_get_upstream_base_url_tcp():
    app = SidecarProxyApp(
        "1.2.2.3:8000",
        max_connections=5000,
        max_keepalive_connections=5,
        keepalive_expiry=60,
        retries=1,
    )
    assert app.get_upstream_base_url({}) == "http://1.2.2.3:8000"


def test_get_upstream_base_url_unix():
    app = SidecarProxyApp(
        "/path/to/proxy.sock",
        max_connections=5000,
        max_keepalive_connections=5,
        keepalive_expiry=60,
        retries=1,
    )
    assert app.get_upstream_base_url({}) == "http://localhost"



import pytest
from pytest_mock import MockerFixture

from mrok.agent.devtools.inspector.server import InspectorServer


def test_server_init():
    server = InspectorServer(port=8901, subscriber_port=18237)
    assert server.command == "python -m mrok.agent.devtools.inspector -p 18237"
    assert server.port == 8901


@pytest.mark.asyncio
async def test_server_on_startup(mocker: MockerFixture):
    console = mocker.MagicMock()
    server = InspectorServer(port=8901, subscriber_port=18237)
    server.console = console

    await server.on_startup(mocker.MagicMock())

    assert console.print.mock_calls[0].args[0] == (
        f"Serving mrok inspector web app on {server.public_url} "
        f"(subscriber port {server.subscriber_port})"
    )
    assert console.print.mock_calls[1].args[0] == "\n[cyan]Press Ctrl+C to quit"
